# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
This module provides several helper minerals/materials.

"""
from __future__ import absolute_import
from __future__ import print_function

from .material import Material, material_property
from .composite import Composite


class HelperRockSwitcher(Material):
    """
    A Helper that represents a Material that switches between different rocks based on a user specified
    select_rock() function based on current temperature and pressure. This class can be used in several
    ways:
    1. By creating an instance and setting select_rock to a lambda that returns a rock
    2. By deriving from this class and implementing select_rock.
    """
    def __init__(self):
        self.current_rock = None
        Material.__init__(self)

    def select_rock(self):
        raise NotImplementedError("Need to implement select_rock() in derived class!")

    def set_method(self, method):
        raise NotImplementedError("Need to implement select_rock() in derived class!")

    def debug_print(self, indent=""):
        print("%sHelperRockSwitcher" % (indent))

    def set_state(self, pressure, temperature):
        Material.set_state(self, pressure, temperature)

        self.current_rock = self.select_rock()
        self.current_rock.set_state(pressure, temperature)

    def unroll(self):
        return self.current_rock.unroll()

    @material_property
    def molar_internal_energy(self):
        return self.current_rock.molar_internal_energy

    @material_property
    def molar_gibbs(self):
        return self.current_rock.molar_gibbs

    @material_property
    def molar_helmholtz(self):
        return self.current_rock.molar_helmholtz

    @material_property
    def molar_mass(self):
        return self.current_rock.molar_mass

    @material_property
    def molar_volume(self):
        return self.current_rock.molar_volume

    @material_property
    def density(self):
        return self.current_rock.density

    @material_property
    def molar_entropy(self):
        return self.current_rock.molar_entropy

    @material_property
    def molar_enthalpy(self):
        return self.current_rock.molar_enthalpy

    @material_property
    def isothermal_bulk_modulus(self):
        return self.current_rock.isothermal_bulk_modulus

    @material_property
    def adiabatic_bulk_modulus(self):
        return self.current_rock.adiabatic_bulk_modulus

    @material_property
    def isothermal_compressibility(self):
        return self.current_rock.isothermal_compressibility

    @material_property
    def adiabatic_compressibility(self):
        return self.current_rock.adiabatic_compressibility

    @material_property
    def shear_modulus(self):
        return self.current_rock.shear_modulus

    @material_property
    def p_wave_velocity(self):
        return self.current_rock.p_wave_velocity

    @material_property
    def bulk_sound_velocity(self):
        return self.current_rock.bulk_sound_velocity

    @material_property
    def shear_wave_velocity(self):
        return self.current_rock.shear_wave_velocity

    @material_property
    def grueneisen_parameter(self):
        return self.current_rock.grueneisen_parameter

    @material_property
    def thermal_expansivity(self):
        return self.current_rock.thermal_expansivity

    @material_property
    def molar_heat_capacity_v(self):
        return self.current_rock.molar_heat_capacity_v

    @material_property
    def molar_heat_capacity_p(self):
        return self.current_rock.molar_heat_capacity_p


class HelperLowHighPressureRockTransition(HelperRockSwitcher):
    """
    A Helper that represents a Material that switches between two given rocks based on a given transition pressure.
    """
    def __init__(self, transition_pressure, low_pressure_rock, high_pressure_rock):
        self.transition_pressure = transition_pressure
        self.rocks = [low_pressure_rock, high_pressure_rock]
        HelperRockSwitcher.__init__(self)
        self._name = "HelperLowHighPressureRockTransition("+str(self.transition_pressure)+ " GPa, " + self.rocks[0].name +  ", " + self.rocks[1].name + ")"

    def select_rock(self):
        if self._pressure < self.transition_pressure:
            return self.rocks[0]
        else:
            return self.rocks[1]

    def set_method(self, method):
        for r in self.rocks:
            r.set_method(method)

    def debug_print(self, indent=""):
        print("%sHelperLowHighPressureRockTransition (%f GPa):" % (indent, self.transition_pressure))
        indent += "  "
        for r in self.rocks:
            r.debug_print(indent)


class HelperSpinTransition(Composite):

    """
    Helper class that makes a mineral that switches between two materials
    (for low and high spin) based on some transition pressure [Pa]
    """

    def __init__(self, transition_pressure, ls_mat, hs_mat):
        """
        Takes a transition pressure, and two minerals.  Use the
        thermoelastic parameters for ls_mat below the transition
        pressure, and the thermoelastic parameters for hs_mat
        above the transition pressure
        """
        Material.__init__(self)
        self.transition_pressure = transition_pressure
        self.ls_mat = ls_mat
        self.hs_mat = hs_mat
        Composite.__init__(self, [ls_mat, hs_mat])

    def debug_print(self, indent=""):
        print("%sHelperSpinTransition:" % indent)
        self.ls_mat.debug_print(indent + "  ")
        self.hs_mat.debug_print(indent + "  ")

    def set_state(self, pressure, temperature):
        if (pressure >= self.transition_pressure):
            Composite.set_fractions(self, [1.0, 0.0])
        else:
            Composite.set_fractions(self, [0.0, 1.0])

        Composite.set_state(self, pressure, temperature)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
JH_2015
^^^^^^^

Solid solutions from Jennings and Holland, 2015 and references therein
(10.1093/petrology/egv020).
The values in this document are all in S.I. units,
unlike those in the original tc file.
"""
import inspect
import numpy as np
from ..classes.solidsolution import SolidSolution
from ..classes.combinedmineral import CombinedMineral
from ..classes.solutionmodel import *
from copy import copy

"""
ENDMEMBERS

Direct import from HP_2011_ds62
"""

from . import HP_2011_ds62

# The next few lines import the classes from HP_2011_ds62,
# so that they can be read as attributes of JH_2015.
# This avoids copying the covariance matrix into the JH_2015,
# as would happen with import HP_2011_ds62
import sys
this_module = sys.modules[__name__]
for m in [m for m in inspect.getmembers(HP_2011_ds62, inspect.isclass)
          if m[1].__module__ == 'burnman.minerals.HP_2011_ds62']:
    setattr(this_module, m[0], m[1])


"""
SOLID SOLUTIONS

The parameters in Jennings and Holland (2015) are given in the following units:
[kJ/mol], [kJ/K/mol], [kJ/kbar/mol]

N.B. The excess entropy terms in these solution models have the opposite sign
to the thermal parameters in Jennings and Holland, 2015.
This is consistent with its treatment as an excess entropy term (W=W_H-T*W_S+P*W_V),
rather than a thermal correction to the interaction parameter   (W=W_0+T*W_T+P*W_P).
"""
class ferropericlase(SolidSolution):
    def __init__(self, molar_fractions=None):
        self.name = 'ferropericlase (FM)'
        self.endmembers = [[HP_2011_ds62.per(), '[Mg]O'],
                           [HP_2011_ds62.fper(), '[Fe]O']]
        self.solution_type = 'symmetric'
        self.energy_interaction = [[18.e3]]
        SolidSolution.__init__(self, molar_fractions=molar_fractions)

class plagioclase(SolidSolution):
    def __init__(self, molar_fractions=None):
        self.name = 'plagioclase (NCAS)'
        self.endmembers = [[HP_2011_ds62.an(), '[Ca][Al]2Si2O8'],
                           [HP_2011_ds62.abh(), '[Na][Al1/2Si1/2]2Si2O8']] # Al-avoidance model
        self.solution_type = 'asymmetric'
        self.alphas = [0.39, 1.]
        self.energy_interaction = [[22.4e3]]
        SolidSolution.__init__(self, molar_fractions=molar_fractions)

class clinopyroxene(SolidSolution):
    def __init__(self, molar_fractions=None):
        self.name = 'clinopyroxene (NCFMASCrO)'
        self.endmembers = [[HP_2011_ds62.di(),   '[Mg][Ca][Si]1/2O6'],
                           [cfs(),  '[Fe][Fe][Si]1/2O6'],
                           [HP_2011_ds62.cats(), '[Al][Ca][Si1/2Al1/2]1/2O6'],
                           [crdi(), '[Cr][Ca][Si1/2Al1/2]1/2O6'],
                           [cess(), '[Fef][Ca][Si1/2Al1/2]1/2O6'],
                           [HP_2011_ds62.jd(),   '[Al][Na][Si]1/2O6'],
                           [cen(),  '[Mg][Mg][Si]1/2O6'],
                           [cfm(),  '[Mg][Fe][Si]1/2O6']] # note cfm ordered endmember
        self.solution_type = 'asymmetric'
        self.alphas = [1.2, 1.0, 1.9, 1.9, 1.9, 1.2, 1.0, 1.0]
        self.energy_interaction = [[20.e3, 12.3e3, 8.e3, 8.e3, 26.e3, 29.8e3, 18.e3],
                                   [25.e3, 34.e3, 34.e3, 24.e3, 7.e3, 4.e3],
                                   [2.e3, 2.e3, 6.e3, 45.7e3, 27.e3],
                                   [2.e3, 3.e3, 48.e3, 36.e3],
                                   [3.e3, 58.e3, 36.e3],
                                   [40.e3, 40.e3],
                                   [4.e3]]
        self.volume_interaction = [[0., -0.1e-5, 0., 0., 0., -0.03e-5, 0.],
                                   [-0.1e-5, 0., 0., 0., 0., 0.],
                                   [0., 0., 0., -0.29e-5, -0.1e-5],
                                   [0., 0., 0., 0.],
                                   [0., 0., 0.],
                                   [0., 0.],
                                   [0.]]
        SolidSolution.__init__(self, molar_fractions=molar_fractions)

class cfs(CombinedMineral):
    def __init__(self):
        CombinedMineral.__init__(self,
                                 name = 'clinoferrosilite',
                                 mineral_list = [HP_2011_ds62.fs()],
                                 molar_amounts = [1.],
                                 free_energy_adjustment=[3.8e3, 3., 0.03e-5])
class crdi(CombinedMineral):
    def __init__(self):
        CombinedMineral.__init__(self,
                                 name = 'chromium diopside',
                                 mineral_list = [HP_2011_ds62.cats(),
                                                 HP_2011_ds62.kos(),
                                                 HP_2011_ds62.jd()],
                                 molar_amounts = [1., 1., -1.],
                                 free_energy_adjustment=[-3.e3, 0., 0.])
class cess(CombinedMineral):
    def __init__(self):
        CombinedMineral.__init__(self,
                                 name = 'ferric diopside',
                                 mineral_list = [HP_2011_ds62.cats(),
                                                 HP_2011_ds62.acm(),
                                                 HP_2011_ds62.jd()],
                                 molar_amounts = [1., 1., -1.],
                                 free_energy_adjustment=[-6.e3, 0., 0.])
class cen(CombinedMineral):
    def __init__(self):
        CombinedMineral.__init__(self,
                                 name = 'clinoenstatite',
                                 mineral_list = [HP_2011_ds62.en()],
                                 molar_amounts = [1.],
                                 free_energy_adjustment=[3.5e3, 2., 0.048e-5])

class cfm(CombinedMineral):
    def __init__(self):
        CombinedMineral.__init__(self,
                                 name = 'ordered clinoferroenstatite',
                                 mineral_list = [HP_2011_ds62.en(),
                                                 HP_2011_ds62.fs()],
                                 molar_amounts = [0.5, 0.5],
                                 free_energy_adjustment=[-3.e3, 0., 0.])

class olivine(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'olivine (FMS)'
        self.endmembers = [[HP_2011_ds62.fo(), '[Mg]2SiO4'],
                           [HP_2011_ds62.fa(), '[Fe]2SiO4']]
        self.solution_type = 'symmetric'
        self.energy_interaction = [[9.e3]]
        SolidSolution.__init__(self, molar_fractions=molar_fractions)


class spinel(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'disordered spinel (CFMASO)'
        self.endmembers = [[HP_2011_ds62.sp(),   '[Al2/3Mg1/3]3O4'],
                           [HP_2011_ds62.herc(), '[Al2/3Fe1/3]3O4'],
                           [HP_2011_ds62.mt(),   '[Fef2/3Fe1/3]3O4'],
                           [HP_2011_ds62.picr(), '[Cr2/3Mg1/3]3O4']]
        self.solution_type = 'symmetric'
        self.energy_interaction = [[4.e3, 56.e3, 39.e3],
                                   [32.e3, 27.e3],
                                   [36.e3]]
        SolidSolution.__init__(self, molar_fractions=molar_fractions)


class garnet(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'garnet (CFMASCrO, low pressure)'
        self.endmembers = [[HP_2011_ds62.py(),   '[Mg]3[Al]2Si3O12'],
                           [HP_2011_ds62.alm(),  '[Fe]3[Al]2Si3O12'],
                           [HP_2011_ds62.gr(),   '[Ca]3[Al]2Si3O12'],
                           [HP_2011_ds62.andr(), '[Ca]3[Fef]2Si3O12'],
                           [HP_2011_ds62.knor(), '[Mg]3[Cr]2Si3O12']]
        self.solution_type = 'symmetric'
        self.energy_interaction = [[4.e3, 35.e3, 91.e3, 2.e3],
                                   [4.e3, 60.e3, 6.e3],
                                   [2.e3, 47.e3],
                                   [101.e3]]
        self.entropy_interaction = [[0., 0., -1.7, 0.],
                                    [0., -1.7, 0.],
                                    [0., 33.8],
                                    [32.1]] # note huge entropy additions! (and sign change from a + bT + cP format)
        self.volume_interaction = [[0.1e-5, 0.1e-5, 0.032e-5, 0.],
                                   [0.1e-5, 0.032e-5, 0.01e-5],
                                   [0., 0.221e-5],
                                   [0.153e-5]]
        SolidSolution.__init__(self, molar_fractions=molar_fractions)

class orthopyroxene(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'orthopyroxene (CFMASCrO)'
        self.endmembers = [[HP_2011_ds62.en(),   '[Mg][Mg][Si]0.5Si1.5O6'],
                           [HP_2011_ds62.fs(),   '[Fe][Fe][Si]0.5Si1.5O6'],
                           [fm(),   '[Fe][Mg][Si]0.5Si1.5O6'],
                           [odi(),  '[Mg][Ca][Si]0.5Si1.5O6'],
                           [HP_2011_ds62.mgts(), '[Al][Mg][Si1/2Al1/2]0.5Si1.5O6'],
                           [cren(), '[Cr][Mg][Si1/2Al1/2]0.5Si1.5O6'],
                           [mess(), '[Fef][Mg][Si1/2Al1/2]0.5Si1.5O6']] # fm ordered phase, fake T-site multiplicity
        self.solution_type = 'asymmetric'
        self.alphas = [1., 1., 1., 1.2, 1., 1., 1.]
        self.energy_interaction = [[5.2e3, 4.e3, 32.2e3, 13.e3, 8.e3, 8.e3],
                                   [4.e3, 24.e3, 7.e3, 10.e3, 10.e3],
                                   [18.e3, 2.e3, 12.e3, 12.e3],
                                   [75.4e3, 30.e3, 30.e3],
                                   [2.e3, 2.e3],
                                   [2.e3]]
        self.volume_interaction = [[0., 0., 0.12e-5, -0.15e-5, 0., 0.],
                                   [0., 0., -0.15e-5, 0., 0.],
                                   [0., -0.15e-5, 0., 0.],
                                   [-0.94e-5, 0., 0.],
                                   [0., 0.],
                                   [0.]]

        SolidSolution.__init__(self, molar_fractions=molar_fractions)



class fm(CombinedMineral):
    def __init__(self):
        CombinedMineral.__init__(self,
                                 name = 'ordered ferroenstatite',
                                 mineral_list = [HP_2011_ds62.en(),
                                                 HP_2011_ds62.fs()],
                                 molar_amounts = [0.5, 0.5],
                                 free_energy_adjustment=[-6.e3, 0., 0.])
class odi(CombinedMineral):
    def __init__(self):
        CombinedMineral.__init__(self,
                                 name = 'orthodiopside',
                                 mineral_list = [HP_2011_ds62.di()],
                                 molar_amounts = [1.],
                                 free_energy_adjustment=[-0.1e3, -0.211, 0.005e-5]) # note sign of *entropy* change.
class cren(CombinedMineral):
    def __init__(self):
        CombinedMineral.__init__(self,
                                 name = 'chromium enstatite',
                                 mineral_list = [HP_2011_ds62.mgts(),
                                                 HP_2011_ds62.kos(),
                                                 HP_2011_ds62.jd()],
                                 molar_amounts = [1., 1., -1.],
                                 free_energy_adjustment=[3.e3, 0., 0.])

class mess(CombinedMineral):
    def __init__(self):
        CombinedMineral.__init__(self,
                                 name = 'ferrienstatite',
                                 mineral_list = [HP_2011_ds62.mgts(),
                                                 HP_2011_ds62.acm(),
                                                 HP_2011_ds62.jd()],
                                 molar_amounts = [1., 1., -1.],
                                 free_energy_adjustment=[-15.e3, 0., 0.15e-5])




def construct_combined_covariance(original_covariance_dictionary,
                                  combined_mineral_list):
    """
    This function takes a dictionary containing a list of endmember_names
    and a covariance_matrix, and a list of CombinedMineral instances, and creates
    an updated covariance dictionary containing those CombinedMinerals

    Parameters
    ----------
    original_covariance_dictionary : dictionary
        Contains a list of strings of endmember_names of length n
        and a 2D numpy array covariance_matrix of shape n x n

    combined_mineral_list : list of instances of :class:`burnman.CombinedMineral`
        List of minerals to be added to the covariance matrix

    Returns
    -------
    cov : dictionary
        Updated covariance dictionary, with the same keys as the original

    """
    cov_orig = original_covariance_dictionary

    # Update names
    cov = {'endmember_names': copy(cov_orig['endmember_names'])}
    for c in combined_mineral_list:
        cov['endmember_names'].append(c.name)

    # Update covariance matrix
    A = np.identity(len(cov_orig['endmember_names']))
    for i, indices in enumerate([[cov_orig['endmember_names'].index(name)
                                  for name in [mbr[0].params['name']
                                               for mbr in c.mixture.endmembers]]
                                 for c in combined_mineral_list]):
        B = np.zeros(len(cov_orig['endmember_names']))
        B[indices] = combined_mineral_list[i].mixture.molar_fractions
        A = np.vstack((A, B))

    cov['covariance_matrix'] = A.dot(cov_orig['covariance_matrix']).dot(A.T)

    return cov


def cov():
    """
    A function which returns the variance-covariance matrix of the zero-point energies
    of all the endmembers in the dataset. Derived from HP_2011_ds62, modified to include all
    the new CombinedMinerals.

    Returns
    -------
    cov : dictionary
        Dictionary keys are:
        - endmember_names: a list of endmember names, and
        - covariance_matrix: a 2D variance-covariance array for the endmember enthalpies of formation
    """
    return construct_combined_covariance(original_covariance_dictionary = HP_2011_ds62.cov(),
                                         combined_mineral_list = [cfs(),
                                                                  crdi(),
                                                                  cess(),
                                                                  cen(),
                                                                  cfm(),
                                                                  fm(),
                                                                  odi(),
                                                                  cren(),
                                                                  mess()])



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Tools
"""

from __future__ import absolute_import

from . import chemistry
from . import eos
from . import equilibration
from . import math
from . import misc
from . import output_seismo
from . import partitioning
from . import polytope
from . import plot
from . import reductions
from . import solution
from . import unitcell



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
import scipy.constants

"""
molar gas constant (R) in J mol^-1 K^-1
"""
gas_constant = scipy.constants.gas_constant


"""
Avogadro constant (N_A) in mol ^ -1
"""
Avogadro = scipy.constants.Avogadro


"""
Boltzmann constant (k_B) in J K^-1.

Note that we are not using scipy.constants.Boltzmann because it is not
available in older versions.
"""
Boltzmann = 1.3806488e-23


"""
Newtonian constant of gravitation (G) in m^3 kg^-1 s^-2
"""

G = scipy.constants.G


"""
Dirac constant (hbar, Planck constant / 2*pi) in J s^-1
"""
Dirac = 1.054571726e-34


"""
1 cm^-1 in J/mol
"""
invcm = 11.9627



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Other minerals
^^^^^^^^^^^^^^

"""
from __future__ import absolute_import

from ..classes import mineral_helpers as helpers
from ..classes.mineral import Mineral
from ..tools.chemistry import dictionarize_formula, formula_mass

from .SLB_2011 import periclase, wuestite, mg_perovskite, fe_perovskite



class liquid_iron( Mineral ):
    """
    Liquid iron equation of state from Anderson and Ahrens (1994)
    """
    def __init__(self):
        formula='Fe1.0'
        formula = dictionarize_formula(formula)
        m = formula_mass(formula)
        rho_0 = 7019.
        V_0 = m/rho_0
        D = 7766.
        Lambda = 1146.
        self.params = {
            'name': 'liquid iron',
            'formula': formula,
            'equation_of_state': 'aa',
            'P_0': 1.e5,
            'T_0': 1811.,
            'S_0': 100., # to fit
            'molar_mass': m,
            'V_0': V_0,
            'E_0': 0.,
            'K_S': 109.7e9,
            'Kprime_S': 4.661,
            'Kprime_prime_S': -0.043e-9,
            'grueneisen_0': 1.735,
            'grueneisen_prime': -0.130/m*1.e-6,
            'grueneisen_n': -1.870,
            'a': [248.92*m, 289.48*m],
            'b': [0.4057*m, -1.1499*m],
            'Theta': [1747.3, 1.537],
            'theta': 5000.,
            'lmda': [-325.23*m, 302.07*m, 30.45*m],
            'xi_0': 282.67*m,
            'F': [D/rho_0, Lambda/rho_0],
            'n': sum(formula.values()),
            'molar_mass': m}
        Mineral.__init__(self)


class ZSB_2013_mg_perovskite(Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'slb3',
            'V_0': 24.45e-6,
            'K_0': 250.5e9,
            'Kprime_0': 4.01,
            'G_0': 172.9e9,
            'Gprime_0': 1.74,
            'molar_mass': .1000,
            'n': 5,
            'Debye_0': 905.9,
            'grueneisen_0': 1.44,
            'q_0': 1.09,
            'eta_s_0': 2.13}  # 2.6

        self.uncertainties = {
            'err_K_0': 3.e9,
            'err_Kprime_0': 0.1,
            'err_G_0': 2.e9,
            'err_Gprime_0': 0.0,
            'err_Debye_0': 5.,
            'err_grueneisen_0': .05,
            'err_q_0': .3,
            'err_eta_s_0': .3}

        Mineral.__init__(self)


class ZSB_2013_fe_perovskite(Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'slb3',
            'V_0': 25.49e-6,
            'K_0': 272.0e9,
            'Kprime_0': 4.1,
            'G_0': 133.0e9,
            'Gprime_0': 1.4,
            'molar_mass': .1319,
            'n': 5,
            'Debye_0': 871.,
            'grueneisen_0': 1.57,
            'q_0': 1.1,
            'eta_s_0': 2.3}  # 2.3

        self.uncertainties = {
            'err_K_0': 40e9,
            'err_Kprime_0': 1.,
            'err_G_0': 40e9,
            'err_Gprime_0': 0.0,
            'err_Debye_0': 26.,
            'err_grueneisen_0': .3,
            'err_q_0': 1.0,
            'err_eta_s_0': 1.0}

        Mineral.__init__(self)


class Speziale_fe_periclase(helpers.HelperSpinTransition):

    def __init__(self):
        helpers.HelperSpinTransition.__init__(
            self, 60.0e9, Speziale_fe_periclase_LS(), Speziale_fe_periclase_HS())
        self.cite = 'Speziale et al. 2007'


class Speziale_fe_periclase_HS(Mineral):

    """
    Speziale et al. 2007, Mg#=83
    """

    def __init__(self):
        self.params = {
            'equation_of_state': 'mgd3',
            'V_0': 22.9e-6,
            'K_0': 157.5e9,
            'Kprime_0': 3.92,
            'molar_mass': .04567,
            'n': 2,
            'Debye_0': 587,
            'grueneisen_0': 1.46,
            'q_0': 1.2}
        Mineral.__init__(self)


class Speziale_fe_periclase_LS(Mineral):

    """
    Speziale et al. 2007, Mg#=83
    """

    def __init__(self):
        self.params = {
            'equation_of_state': 'mgd3',
            'V_0': 21.49e-6,
            'K_0': 186.0e9,
            'Kprime_0': 4.6,
            'molar_mass': .04567,
            'n': 2,
            'Debye_0': 587.,
            'grueneisen_0': 1.46,
            'q_0': 1.2}
        Mineral.__init__(self)


class Liquid_Fe_Anderson(Mineral):

    """
    Anderson & Ahrens, 1994 JGR
    """

    def __init__(self):
        self.params = {
            'equation_of_state': 'bm4',
            'V_0': 7.95626e-6,
            'K_0': 109.7e9,
            'Kprime_0': 4.66,
            'Kprime_prime_0': -0.043e-9,
            'molar_mass': 0.055845,
        }
        Mineral.__init__(self)


class Fe_Dewaele(Mineral):

    """
    Dewaele et al., 2006, Physical Review Letters
    """

    def __init__(self):
        self.params = {
            'equation_of_state': 'vinet',
            'V_0': 6.75e-6,
            'K_0': 163.4e9,
            'Kprime_0': 5.38,
            'molar_mass': 0.055845,
            'n': 1}
        Mineral.__init__(self)



from __future__ import print_function
# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2018 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
from scipy.optimize import nnls

from ..tools.chemistry import dictionarize_formula, formula_mass
from ..tools.misc import OrderedCounter


def composition_property(func):
    """
    Decorator @composition_property to be used for cached properties of compositions.

    To be used on function in Composition or derived classes that should be exposed
    as read-only properties that are cached.
    """
    class mat_obj():

        def __init__(self, func):
            self.func = func
            self.varname = self.func.__name__

        def get(self, obj):
            if not hasattr(obj, "_cached"):
                raise Exception("The composition_property decorator could not find class member _cached. "
                                "Did you forget to call Composition.__init__(self) in __init___?")
            cache_array = getattr(obj, "_cached")
            if self.varname not in cache_array:
                cache_array[self.varname] = self.func(obj)
            return cache_array[self.varname]

    return property(mat_obj(func).get, doc=func.__doc__)


def file_to_composition_list(fname, unit_type, normalize):
    """
    Takes an input file with a specific format and returns a list of compositions
    (and associated comments) contained in that file.

    Parameters
    ----------
    fname : string
        Path to ascii file containing composition data.
        Lines beginning with a hash are not read.
        The first read-line of the datafile contains a list of tab or
        space-separated components (e.g. FeO or SiO2), followed by the word Comment.
        Following lines are lists of floats with the amounts of each component.
        After the component amounts, the user can write anything they like
        in the Comment section.
    unit_type : 'weight' or 'molar'
        Specify whether the compositions in the file are given as weight or
        molar amounts.
    normalize : boolean
        If False, absolute numbers of moles/grams of component are stored,
        otherwise the component amounts of returned compositions will
        sum to one (until Composition.renormalize() is used).

    """
    lines = list(filter(None, [line.rstrip('\n').split() for line in open(fname) if line[0] != '#']))
    n_components = lines[0].index("Comment")
    components = lines[0][:n_components]
    comments = [line[n_components:] for line in lines[1:]]
    compositions = np.array([map(float, l) for l in list(zip(*(list(zip(*lines[1:]))[:n_components])))])
    return [Composition(OrderedCounter(dict(zip(components, c))), unit_type, normalize) for c in compositions], comments


class Composition(object):
    """
    Class for a composition object, which can be used
    to store, modify and renormalize compositions,
    and also convert between molar, weight,
    and atomic amounts.

    This class is available as ``burnman.Composition``.
    """

    def __init__(self, composition_dictionary, unit_type='weight', normalize=False):
        """
        Create a composition using a dictionary and unit type.

        Parameters
        ----------
        composition_dictionary : dictionary
            Dictionary of components (given as a string) and their amounts.
        unit_type : 'weight' or 'molar' (optional, 'weight' as standard)
            Specify whether the input composition is given as weight or
            molar amounts.
        normalize : boolean
            If False, absolute numbers of moles/grams of component are stored,
            otherwise the component amounts of returned compositions will
            sum to one (until Composition.renormalize() is used).
        """

        self._cached = {}

        n_total = float(sum(composition_dictionary.values()))

        normalized_dictionary = {}
        for k in composition_dictionary.keys():
            normalized_dictionary[k] = composition_dictionary[k]/n_total

        self.normalization_component = {'weight': 'total',
                                        'molar': 'total',
                                        'atomic': 'total'}

        self.normalization_amount = {'weight': 1.,
                                     'molar': 1.,
                                     'atomic': 1.}


        # component formulae
        self.component_formulae = {c: dictionarize_formula(c)
                                   for c in composition_dictionary.keys()}

        # elemental compositions of components
        self.element_list = OrderedCounter()
        for component in self.component_formulae.values():
            self.element_list += OrderedCounter({element: n_atoms
                                          for (element, n_atoms) in component.items()})
        self.element_list = list(self.element_list.keys())

        if unit_type == 'weight':
            if normalize:
                self._cached['weight_composition'] = OrderedCounter(normalized_dictionary)
            else:
                self._cached['weight_composition'] = OrderedCounter(composition_dictionary)

                mole_total = sum([composition_dictionary[c] /
                                  formula_mass(self.component_formulae[c])
                                  for c in composition_dictionary.keys()])
                self.normalization_amount['weight'] = n_total
                self.normalization_amount['molar'] = mole_total
                self.normalization_amount['atomic'] = sum(self._moles_component_to_atoms(self.molar_composition).values())

        elif unit_type == 'molar':
            if normalize:
                self._cached['molar_composition'] = OrderedCounter(normalized_dictionary)
            else:
                self._cached['molar_composition'] = OrderedCounter(composition_dictionary)

                weight_total = sum([composition_dictionary[c] *
                                    formula_mass(self.component_formulae[c])
                                    for c in composition_dictionary.keys()])

                self.normalization_amount['weight'] = weight_total
                self.normalization_amount['molar'] = n_total
                self.normalization_amount['atomic'] = sum(self._moles_component_to_atoms(self.molar_composition).values())

        else:
            raise Exception('Unit type not yet implemented. '
                            'Should be either weight or molar.')





    def renormalize(self, unit_type, normalization_component, normalization_amount):
        """
        Change the normalization for a given unit type
        (weight, molar, or atomic)
        Resets cached composition only for that unit type

        Parameters
        ----------
        unit_type : 'weight', 'molar' or 'atomic'
            Unit type composition to be renormalised
        normalization_component: string
            Component/element on which to renormalize.
            String must either be one of the components/elements
            already in composite, or have the value 'total'
        normalization_amount: float
            Amount of component in the renormalised composition
        """
        if unit_type == 'weight':
            s = self.weight_composition
        elif unit_type == 'molar':
            s = self.molar_composition
        elif unit_type == 'atomic':
            s = self.atomic_composition
        else:
            raise Exception('Unit type not recognised. '
                            'Should be one of weight, molar and atomic')

        self.normalization_component[unit_type] = normalization_component
        self.normalization_amount[unit_type] = normalization_amount
        self._cached[unit_type+'_composition'] = self._normalize_to_basis(s, unit_type)

    def add_components(self, composition_dictionary, unit_type):
        """
        Add (or remove) components from the composition.
        The components are added to the current state of the
        (weight or molar) composition; if the composition has
        been renormalised, then this should be taken into account.

        Parameters
        ----------
        composition_dictionary : dictionary
            Components to add, and their amounts, in dictionary form
        unit_type : 'weight' or 'molar'
            Unit type of the components to be added
        """
        if unit_type == 'weight':
            composition = self.weight_composition
        elif unit_type == 'molar':
            composition = self.molar_composition
        else:
            raise Exception('Unit type not recognised. '
                            'Should be either weight or molar.')

        composition += OrderedCounter(composition_dictionary)

        self.__init__(composition, unit_type)

    def change_component_set(self, new_component_list):
        """
        Change the set of basis components without
        changing the bulk composition.

        Will raise an exception if the new component set is
        invalid for the given composition.

        Parameters
        ----------
        new_component_list : list of strings
            New set of basis components.
        """
        composition = np.array([self.atomic_composition[element]
                                for element in self.element_list])
        component_matrix = np.zeros((len(new_component_list), len(self.element_list)))

        for i, component in enumerate(new_component_list):
            formula = dictionarize_formula(component)
            for element, n_atoms in formula.items():
                component_matrix[i][self.element_list.index(element)] = n_atoms

        sol = nnls(component_matrix.T, composition)
        if sol[1] < 1.e-12:
            component_amounts = sol[0]
        else:
            raise Exception('Failed to change component set. '
                            'Could not find a non-negative least squares solution. '
                            'Can the bulk composition be described with this set of components?')

        composition = OrderedCounter(dict(zip(new_component_list, component_amounts)))
        self.__init__(composition, 'molar')

    def _normalize_to_basis(self, composition, unit_type):
        if self.normalization_component[unit_type] == 'total':
            n_orig = float(sum(composition.values()))
        else:
            n_orig = composition[self.normalization_component[unit_type]]

        for k in composition.keys():
            composition[k] *= self.normalization_amount[unit_type]/n_orig

        return composition

    @composition_property
    def molar_composition(self):
        """
        Returns the molar composition as a counter [moles]
        """
        mole_compositions = OrderedCounter({c: self.weight_composition[c] /
                                            formula_mass(self.component_formulae[c])
                                            for c in self.weight_composition.keys()})

        return self._normalize_to_basis(mole_compositions, 'molar')

    @composition_property
    def weight_composition(self):
        """
        Returns the weight composition as a counter [g]
        """
        weight_compositions = OrderedCounter({c: self.molar_composition[c] *
                                              formula_mass(self.component_formulae[c])
                                              for c in self.molar_composition.keys()})

        return self._normalize_to_basis(weight_compositions, 'weight')

    @composition_property
    def atomic_composition(self):
        """
        Returns the atomic composition as a counter [moles]
        """
        atom_compositions = self._moles_component_to_atoms(self.molar_composition)

        return self._normalize_to_basis(atom_compositions, 'atomic')

    def _moles_component_to_atoms(self, molar_composition_dictionary):
        component_matrix = np.zeros((len(self.component_formulae), len(self.element_list)))
        molar_composition_vector = np.zeros(len(self.component_formulae))
        for i, (component, formula) in enumerate(self.component_formulae.items()):
            molar_composition_vector[i] = molar_composition_dictionary[component]

            for element, n_atoms in formula.items():
                component_matrix[i][self.element_list.index(element)] = n_atoms

        atom_compositions = np.dot(molar_composition_vector, component_matrix)
        return OrderedCounter(dict(zip(self.element_list, atom_compositions)))


    def print(self, unit_type, significant_figures=1,
              normalization_component='total', normalization_amount=100.):
        """
        Pretty-print function for the composition
        This does not renormalize the Composition internally

        Parameters
        ----------
        unit_type : 'weight', 'molar' or 'atomic'
            Unit type in which to print the composition
        significant_figures : integer
            Number of significant figures for each amount
        normalization_component: string
            Component/element on which to renormalize.
            String must either be one of the components/elements
            already in composite, or have the value 'total'.
            (default = 'total')
        normalization_amount: float
            Amount of component in the renormalised composition.
            (default = '100.')
        """
        if unit_type == 'weight':
            print('Weight composition')

            if normalization_component == 'total':
                total_stored = float(sum(self.weight_composition.values()))
            else:
                total_stored = self.weight_composition[normalization_component]
            f = normalization_amount/total_stored

            for (key, value) in sorted(self.weight_composition.items()):
                print('{0}: {1:0.{sf}f}'.format(key, value*f, sf=significant_figures))
        elif unit_type == 'molar':
            print('Molar composition')

            if normalization_component == 'total':
                total_stored = float(sum(self.molar_composition.values()))
            else:
                total_stored = self.molar_composition[normalization_component]
            f = normalization_amount/total_stored

            for (key, value) in sorted(self.molar_composition.items()):
                print('{0}: {1:0.{sf}f}'.format(key, value*f, sf=significant_figures))
        elif unit_type == 'atomic':
            print('Atomic composition')

            if normalization_component == 'total':
                total_stored = float(sum(self.atomic_composition.values()))
            else:
                total_stored = self.atomic_composition[normalization_component]
            f = normalization_amount/total_stored

            for (key, value) in sorted(self.atomic_composition.items()):
                print('{0}: {1:0.{sf}f}'.format(key, value*f, sf=significant_figures))
        else:
            raise Exception('unit_type not yet implemented. Should be either weight,  molar or atomic.')



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import numpy as np
import scipy.optimize as opt
import warnings

# Try to import the jit from numba.  If it is
# not available, just go with the standard
# python interpreter
try:
    from numba import jit
except ImportError:
    def jit(fn):
        return fn


from . import birch_murnaghan as bm
from . import debye
from . import equation_of_state as eos
from ..tools.math import bracket


@jit
def _grueneisen_parameter_fast(V_0, volume, gruen_0, q_0):
    """global function with plain parameters so jit will work"""
    x = V_0 / volume
    f = 1. / 2. * (pow(x, 2. / 3.) - 1.)
    a1_ii = 6. * gruen_0  # EQ 47
    a2_iikk = -12. * gruen_0 + 36. * \
        gruen_0 * gruen_0 - 18. * q_0 * gruen_0  # EQ 47
    nu_o_nu0_sq = 1. + a1_ii * f + (1. / 2.) * a2_iikk * f * f  # EQ 41
    return 1. / 6. / nu_o_nu0_sq * (2. * f + 1.) * (a1_ii + a2_iikk * f)


@jit
def _delta_pressure(x, pressure, temperature, V_0, T_0, Debye_0, n, a1_ii, a2_iikk, b_iikk, b_iikkmm):

    f = 0.5 * (pow(V_0 / x, 2. / 3.) - 1.)
    debye_temperature = Debye_0 * np.sqrt(1. + a1_ii * f + 1. / 2. * a2_iikk * f * f)
    E_th = debye.thermal_energy(
        temperature, debye_temperature, n)  # thermal energy at temperature T
    E_th_ref = debye.thermal_energy(
        T_0, debye_temperature, n)  # thermal energy at reference temperature
    nu_o_nu0_sq = 1. + a1_ii * f + (1. / 2.) * a2_iikk * f * f  # EQ 41
    gr = 1. / 6. / nu_o_nu0_sq * (2. * f + 1.) * (a1_ii + a2_iikk * f)

    return (1. / 3.) * (pow(1. + 2. * f, 5. / 2.)) * ((b_iikk * f) + (0.5 * b_iikkmm * f * f)) \
        + gr * (E_th - E_th_ref) / x - pressure  # EQ 21


class SLBBase(eos.EquationOfState):

    """
    Base class for the finite strain-Mie-Grueneiesen-Debye equation of state detailed
    in :cite:`Stixrude2005`.  For the most part the equations are
    all third order in strain, but see further the :class:`burnman.slb.SLB2` and
    :class:`burnman.slb.SLB3` classes.
    """

    def _debye_temperature(self, x, params):
        """
        Finite strain approximation for Debye Temperature [K]
        x = ref_vol/vol
        """
        f = 1. / 2. * (pow(x, 2. / 3.) - 1.)
        a1_ii = 6. * params['grueneisen_0']  # EQ 47
        a2_iikk = -12. * params['grueneisen_0'] + 36. * pow(
            params['grueneisen_0'], 2.) - 18. * params['q_0'] * params['grueneisen_0']  # EQ 47
        return params['Debye_0'] * np.sqrt(1. + a1_ii * f + 1. / 2. * a2_iikk * f * f)

    def volume_dependent_q(self, x, params):
        """
        Finite strain approximation for :math:`q`, the isotropic volume strain
        derivative of the grueneisen parameter.
        """
        f = 1. / 2. * (pow(x, 2. / 3.) - 1.)
        a1_ii = 6. * params['grueneisen_0']  # EQ 47
        a2_iikk = -12. * params['grueneisen_0'] + 36. * pow(
            params['grueneisen_0'], 2.) - 18. * params['q_0'] * params['grueneisen_0']  # EQ 47
        nu_o_nu0_sq = 1. + a1_ii * f + (1. / 2.) * a2_iikk * f * f  # EQ 41
        gr = 1. / 6. / nu_o_nu0_sq * (2. * f + 1.) * (a1_ii + a2_iikk * f)
        if np.abs(params['grueneisen_0']) < 1.e-10:  # avoids divide by zero if grueneisen_0 = 0.
            q = 1. / 9. * (18. * gr - 6.)
        else:
            q = 1. / 9. * \
                (18. * gr - 6. - 1. / 2. / nu_o_nu0_sq *
                 (2. * f + 1.) * (2. * f + 1.) * a2_iikk / gr)
        return q

    def _isotropic_eta_s(self, x, params):
        """
        Finite strain approximation for :math:`eta_{s0}`, the isotropic shear
        strain derivative of the grueneisen parameter.
        """
        f = 1. / 2. * (pow(x, 2. / 3.) - 1.)
        a2_s = -2. * params['grueneisen_0'] - 2. * params['eta_s_0']  # EQ 47
        a1_ii = 6. * params['grueneisen_0']  # EQ 47
        a2_iikk = -12. * params['grueneisen_0'] + 36. * pow(
            params['grueneisen_0'], 2.) - 18. * params['q_0'] * params['grueneisen_0']  # EQ 47
        nu_o_nu0_sq = 1. + a1_ii * f + \
            (1. / 2.) * a2_iikk * pow(f, 2.)  # EQ 41
        gr = 1. / 6. / nu_o_nu0_sq * (2. * f + 1.) * (a1_ii + a2_iikk * f)
        # EQ 46 NOTE the typo from Stixrude 2005:
        eta_s = - gr - \
            (1. / 2. * pow(nu_o_nu0_sq, -1.) * pow((2. * f) + 1., 2.) * a2_s)

        return eta_s

    # calculate isotropic thermal pressure, see
    # Matas et. al. (2007) eq B4
    def _thermal_pressure(self, T, V, params):
        Debye_T = self._debye_temperature(params['V_0'] / V, params)
        gr = self.grueneisen_parameter(0., T, V, params)  # P not important
        P_th = gr * debye.thermal_energy(T, Debye_T, params['n']) / V
        return P_th

    def volume(self, pressure, temperature, params):
        """
        Returns molar volume. :math:`[m^3]`
        """
        T_0 = params['T_0']
        Debye_0 = params['Debye_0']
        V_0 = params['V_0']
        n = params['n']

        a1_ii = 6. * params['grueneisen_0']  # EQ 47
        a2_iikk = -12. * params['grueneisen_0'] + 36. * pow(
            params['grueneisen_0'], 2.) - 18. * params['q_0'] * params['grueneisen_0']  # EQ 47

        b_iikk = 9. * params['K_0']  # EQ 28
        b_iikkmm = 27. * params['K_0'] * (params['Kprime_0'] - 4.)  # EQ 29z

        # we need to have a sign change in [a,b] to find a zero. Let us start with a
        # conservative guess:
        args = (pressure, temperature, V_0, T_0,
                Debye_0, n, a1_ii, a2_iikk, b_iikk, b_iikkmm)
        try:
            sol = bracket(_delta_pressure, params[
                          'V_0'], 1.e-2 * params['V_0'], args)
        except ValueError:
            raise Exception(
                'Cannot find a volume, perhaps you are outside of the range of validity for the equation of state?')
        return opt.brentq(_delta_pressure, sol[0], sol[1], args=args)

    def pressure(self, temperature, volume, params):
        """
        Returns the pressure of the mineral at a given temperature and volume [Pa]
        """
        debye_T = self._debye_temperature(params['V_0'] / volume, params)
        gr = self.grueneisen_parameter(
            0.0, temperature, volume, params)  # does not depend on pressure
        E_th = debye.thermal_energy(temperature, debye_T, params['n'])
        E_th_ref = debye.thermal_energy(
            params['T_0'], debye_T, params['n'])  # thermal energy at reference temperature

        b_iikk = 9. * params['K_0']  # EQ 28
        b_iikkmm = 27. * params['K_0'] * (params['Kprime_0'] - 4.)  # EQ 29
        f = 0.5 * (pow(params['V_0'] / volume, 2. / 3.) - 1.)  # EQ 24
        P = (1. / 3.) * (pow(1. + 2. * f, 5. / 2.)) \
            * ((b_iikk * f) + (0.5 * b_iikkmm * pow(f, 2.)))\
            + gr * (E_th - E_th_ref) / volume  # EQ 21

        return P

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter :math:`[unitless]`
        """
        return _grueneisen_parameter_fast(params['V_0'], volume, params['grueneisen_0'], params['q_0'])

    def isothermal_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`[Pa]`
        """
        T_0 = params['T_0']
        debye_T = self._debye_temperature(params['V_0'] / volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)

        E_th = debye.thermal_energy(
            temperature, debye_T, params['n'])  # thermal energy at temperature T
        E_th_ref = debye.thermal_energy(
            T_0, debye_T, params['n'])  # thermal energy at reference temperature

        C_v = debye.molar_heat_capacity_v(
            temperature, debye_T, params['n'])  # heat capacity at temperature T
        C_v_ref = debye.molar_heat_capacity_v(
            T_0, debye_T, params['n'])  # heat capacity at reference temperature

        q = self.volume_dependent_q(params['V_0'] / volume, params)

        K = bm.bulk_modulus(volume, params) \
            + (gr + 1. - q) * (gr / volume) * (E_th - E_th_ref) \
            - (pow(gr, 2.) / volume) * (C_v * temperature - C_v_ref * T_0)

        return K

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus. :math:`[Pa]`
        """
        K_T = self.isothermal_bulk_modulus(
            pressure, temperature, volume, params)
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)
        K_S = K_T * (1. + gr * alpha * temperature)
        return K_S

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus. :math:`[Pa]`
        """
        T_0 = params['T_0']
        debye_T = self._debye_temperature(params['V_0'] / volume, params)
        eta_s = self._isotropic_eta_s(params['V_0'] / volume, params)

        E_th = debye.thermal_energy(temperature, debye_T, params['n'])
        E_th_ref = debye.thermal_energy(T_0, debye_T, params['n'])

        if self.order == 2:
            return bm.shear_modulus_second_order(volume, params) - eta_s * (E_th - E_th_ref) / volume
        elif self.order == 3:
            return bm.shear_modulus_third_order(volume, params) - eta_s * (E_th - E_th_ref) / volume
        else:
            raise NotImplementedError("")

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume. :math:`[J/K/mol]`
        """
        debye_T = self._debye_temperature(params['V_0'] / volume, params)
        return debye.molar_heat_capacity_v(temperature, debye_T, params['n'])

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant pressure. :math:`[J/K/mol]`
        """
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)
        C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        C_p = C_v * (1. + gr * alpha * temperature)
        return C_p

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns thermal expansivity. :math:`[1/K]`
        """
        C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)
        K = self.isothermal_bulk_modulus(pressure, temperature, volume, params)
        alpha = gr * C_v / K / volume
        return alpha

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy at the pressure and temperature of the mineral [J/mol]
        """
        G = self.helmholtz_free_energy(
            pressure, temperature, volume, params) + pressure * volume
        return G

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy at the pressure and temperature of the mineral [J/mol]
        """
        return self.helmholtz_free_energy(pressure, temperature, volume, params) + \
            temperature * \
            self.entropy(pressure, temperature, volume, params)

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the entropy at the pressure and temperature of the mineral [J/K/mol]
        """
        Debye_T = self._debye_temperature(params['V_0'] / volume, params)
        S = debye.entropy(temperature, Debye_T, params['n'])
        return S

    def enthalpy(self, pressure, temperature, volume, params):
        """
        Returns the enthalpy at the pressure and temperature of the mineral [J/mol]
        """

        return self.helmholtz_free_energy(pressure, temperature, volume, params) + \
            temperature * self.entropy(pressure, temperature, volume, params) + \
            pressure * volume

    def helmholtz_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Helmholtz free energy at the pressure and temperature of the mineral [J/mol]
        """
        x = params['V_0'] / volume
        f = 1. / 2. * (pow(x, 2. / 3.) - 1.)
        Debye_T = self._debye_temperature(params['V_0'] / volume, params)

        F_quasiharmonic = debye.helmholtz_free_energy(temperature, Debye_T, params['n']) - \
            debye.helmholtz_free_energy(
                params['T_0'], Debye_T, params['n'])

        b_iikk = 9. * params['K_0']  # EQ 28
        b_iikkmm = 27. * params['K_0'] * (params['Kprime_0'] - 4.)  # EQ 29

        F = params['F_0'] + \
            0.5 * b_iikk * f * f * params['V_0'] + (1. / 6.) * params['V_0'] * b_iikkmm * f * f * f +\
            F_quasiharmonic

        return F

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """
        if 'T_0' not in params:
            params['T_0'] = 300.

        # If eta_s_0 is not included this is presumably deliberate,
        # as we can model density and bulk modulus just fine without it,
        # so just add it to the dictionary as nan
        # The same goes for the standard state Helmholtz free energy
        if 'eta_s_0' not in params:
            params['eta_s_0'] = float('nan')
        if 'F_0' not in params:
            params['F_0'] = float('nan')

        # First, let's check the EoS parameters for Tref
        bm.BirchMurnaghanBase.validate_parameters(
            bm.BirchMurnaghanBase(), params)

        # Now check all the required keys for the
        # thermal part of the EoS are in the dictionary
        expected_keys = [
            'molar_mass', 'n', 'Debye_0', 'grueneisen_0', 'q_0', 'eta_s_0']
        for k in expected_keys:
            if k not in params:
                raise KeyError('params object missing parameter : ' + k)

        # Finally, check that the values are reasonable.
        if params['T_0'] < 0.:
            warnings.warn('Unusual value for T_0', stacklevel=2)
        if params['molar_mass'] < 0.001 or params['molar_mass'] > 10.:
            warnings.warn('Unusual value for molar_mass', stacklevel=2)
        if params['n'] < 1. or params['n'] > 1000.:
            warnings.warn('Unusual value for n', stacklevel=2)
        if params['Debye_0'] < 1. or params['Debye_0'] > 10000.:
            warnings.warn('Unusual value for Debye_0', stacklevel=2)
        if params['grueneisen_0'] < -0.005 or params['grueneisen_0'] > 10.:
            warnings.warn('Unusual value for grueneisen_0', stacklevel=2)
        if params['q_0'] < -10. or params['q_0'] > 10.:
            warnings.warn('Unusual value for q_0', stacklevel=2)
        if params['eta_s_0'] < -10. or params['eta_s_0'] > 10.:
            warnings.warn('Unusual value for eta_s_0', stacklevel=2)


class SLB3(SLBBase):

    """
    SLB equation of state with third order finite strain expansion for the
    shear modulus (this should be preferred, as it is more thermodynamically
    consistent.)
    """

    def __init__(self):
        self.order = 3


class SLB2(SLBBase):

    """
    SLB equation of state with second order finite strain expansion for the
    shear modulus.  In general, this should not be used, but sometimes
    shear modulus data is fit to a second order equation of state.  In that
    case, you should use this.  The moral is, be careful!
    """

    def __init__(self):
        self.order = 2



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.


from __future__ import absolute_import

import numpy as np
import matplotlib.pyplot as plt
import pkgutil

from ..classes.planet import Planet
from ..classes.layer import Layer


def write_tvel_file(planet_or_layer, modelname='burnmanmodel',
                    background_model=None):
    """
    Function to write input file for obspy travel time calculations.
    Note: Because density isn't defined for most 1D seismic models, densities
    are output as zeroes.  The tvel format has a column for density,
    but this column is not used by obspy for travel time calculations.
    Parameters
    ----------
    planet_or_layer  :  burnman.Planet() or burnman.Layer()
        Planet or layer to write out to tvel file
    filename : string
        Filename to read to
    background_model : burnman.seismic.Seismic1DModel()
        1D seismic model to fill in parts of planet
        (likely to be an earth model) that aren't defined by layer
        (only need when using Layer())
    """

    if not isinstance(planet_or_layer, (Planet, Layer)):
        raise TypeError("Input must be a Planet() or Layer() ")

    if isinstance(planet_or_layer, Layer):
        assert(background_model)
        layer = planet_or_layer
        depths = background_model.internal_depth_list()

        above_layer = np.where(
            depths < (np.max(depths) - layer.outer_radius))[-1]
        below_layer = np.where(
            depths > (np.max(depths) - layer.inner_radius))[0]

        data_above = list(zip(depths[above_layer] / 1.e3,
                              background_model.v_p(depths[above_layer]) / 1.e3,
                              background_model.v_s(depths[above_layer]) / 1.e3,
                              np.zeros_like(depths[above_layer])))
        data_layer = list(zip((np.max(depths) - layer.radii)[::- 1] / 1.e3,
                              layer.v_p[::-1] / 1.e3,
                              layer.v_s[::-1] / 1.e3,
                              layer.density[::-1] / 1.e3))
        data_below = list(zip(depths[below_layer] / 1.e3,
                              background_model.v_p(depths[below_layer]) / 1.e3,
                              background_model.v_s(depths[below_layer]) / 1.e3,
                              np.zeros_like(depths[below_layer])))

        data = data_above + data_layer + data_below

        header = (f'{layer.name}  model from BurnMan between a radius of '
                  f'{str(layer.inner_radius)} and '
                  f'{str(layer.outer_radius)} km \n'
                  f'{background_model.__class__.__name__} '
                  f'for the rest of the Earth')
        with open(modelname + '.tvel', 'wb') as f:
            np.savetxt(f, data, header=header, fmt='%5.2f', delimiter='\t')

    if isinstance(planet_or_layer, Planet):
        planet = planet_or_layer
        data = list(zip((planet.radius_planet - planet.radii)[::-1] / 1.e3,
                        planet.v_p[::-1] / 1.e3,
                        planet.v_s[::-1] / 1.e3,
                        planet.density[::- 1] / 1.e3))

        header = (f'{planet.name} model from BurnMan with a radius of '
                  f'{str(planet.radius_planet)} km \n'
                  f'Layers of planet are '
                  f'{", ".join(layer.name for layer in planet.layers)}')
        with open(modelname + '.tvel', 'wb') as f:
            np.savetxt(f, data, header=header, fmt='%5.2f', delimiter='\t')


def write_axisem_input(layers, modelname='burnmanmodel_foraxisem',
                       axisem_ref='axisem_prem_ani_noocean.txt',
                       plotting=False):
    """
    Writing velocities and densities to AXISEM (www.axisem.info) input file.
    The input can be a single layer, or a list of layers taken from a planet (planet.layers).
    Currently this function will implement explicit discontinuities between layers in the seismic model.
    Currently this function is only set for Earth.


    Parameters
    ----------
    layers : list of one or more burnman.Layer()
        List of layers to put in axisem file
    modelname : string
        Name of model, appears in name of output file
    axisem_ref : string
        Reference file, used to copy the header and for the rest of the planet, in the case of a Layer(), default = 'axisem_prem_ani_noocean.txt'
    plotting: Boolean
        True means plot of the old model and replaced model will be shown (default = False)
    """

    if not isinstance(layers[0], Layer):
        raise TypeError("Input must be a list of Layer()")
    # Load reference input
    datastream = pkgutil.get_data(
        'burnman', 'data/input_seismic/' + axisem_ref)
    lines = [line.strip()
             for line in datastream.decode('ascii').split('\n') if line.strip()]
    table = []
    for line in lines[18:]:
        numbers = np.fromstring(line, sep=' ')
        if len(numbers) > 0:
            if line[0] != "#" and line[0] != "%":
                table.append(numbers)
    table = np.array(table)
    # format is
    # radius density vpv vsv Qk Qmu vph vsh eta

    if plotting:
        plt.figure(figsize=(12, 6))
        plt.plot(table[:, 0] / 1.e3, table[:, 2] / 1.e3,
                 color='g', linestyle='--')
        plt.plot(table[:, 0] / 1.e3, table[:, 3] / 1.e3,
                 color='b', linestyle='--')
        plt.plot(table[:, 0] / 1.e3, table[:, 1] / 1.e3,
                 color='r', linestyle='--')

    for layer in layers:
        # Cutting out range to input in Axisem reference file, and adding in
        # model values at the top and bottom of Layer.
        i_min = next(x[0] for x in enumerate(table[:, 0])
                     if x[1] <= layer.outer_radius)
        if table[i_min, 0] - layer.outer_radius < 0:
            table = np.insert(table, i_min, None, axis=0)
            table[i_min, 0] = layer.outer_radius

        i_max = next(x[0] for x in enumerate(table[:, 0])
                     if x[1] <= layer.inner_radius)

        if table[i_max, 0] - layer.inner_radius < 0:
            table = np.insert(table, i_max, None, axis=0)
            table[i_max, 0] = layer.inner_radius

        layer_vp, layer_vs, layer_density = layer.evaluate(
            ['v_p', 'v_s', 'density'], radlist=table[i_min:i_max, 0], radius_planet=np.max(table[:, 0]))

        table[i_min:i_max, 2] = layer_vp
        table[i_min:i_max, 6] = layer_vp
        table[i_min:i_max, 3] = layer_vs
        table[i_min:i_max, 7] = layer_vs
        table[i_min:i_max, 1] = layer_density

    # WRITE OUT FILE
    filename = 'axisem_' + modelname + '.txt'
    f = open(filename, 'w')
    print('Writing ' + filename + ' ...')
    f.write(f'# Input file {modelname} for AXISEM created using BurnMan, '
            f'replacing {axisem_ref} between '
            f'{str(np.round(layer.inner_radius / 1.e3))} and '
            f'{str(np.round(layer.outer_radius / 1.e3))} km\n')

    discontinuity = 0  # Number discontinuities
    f.write('NAME ' + modelname + '\n')
    for line in lines[2:18]:
        f.write(line[:] + '\n')
    for i in range(len(table[:, 0])):
        if i > 0 and table[i, 0] == table[i - 1, 0]:
            discontinuity = discontinuity + 1
            f.write(f'#          Discontinuity   {str(discontinuity)}, '
                    f'depth:    {str(np.round((6371.e3 - table[i, 0]) / 1.e3, decimals=2))} km \n')

        f.write(f'{table[i, 0]:8.0f} {table[i, 1]:9.2f} {table[i, 2]:9.2f} '
                f'{table[i, 3]:9.2f} {table[i, 4]:9.1f} {table[i, 5]:9.1f} '
                f'{table[i, 6]:9.2f} {table[i, 7]:9.2f} {table[i, 8]:9.5f} \n')

    f.close()

    if plotting:

        plt.plot(table[:, 0] / 1.e3, table[:, 2] / 1.e3,
                 color='g', linestyle='-', label='V_p')
        plt.plot(table[:, 0] / 1.e3, table[:, 3] / 1.e3,
                 color='b', linestyle='-', label='V_s')
        plt.plot(table[:, 0] / 1.e3, table[:, 1] / 1.e3,
                 color='r', linestyle='-', label='density')

        plt.title(f'{filename} = {axisem_ref} replaced between '
                  f'{str(layer.inner_radius / 1.e3)} and '
                  f'{str(layer.outer_radius / 1.e3)} km')
        plt.legend(loc='lower right')
        plt.show()


def write_mineos_input(layers, modelname='burnmanmodel_formineos',
                       mineos_ref='mineos_prem_noocean.txt', plotting=False):
    """
    Writing velocities and densities to
    Mineos (https://geodynamics.org/cig/software/mineos/) input file
    Note:
        - Currently, this function only honors the discontinuities already
        in the synthetic input file, so it is best to only replace
        certain layers with burnman values

    Parameters
    ----------
    layers : list of one or more burnman.Layer()
        List of layers to put in axisem file
    modelname : string
        Name of model, appears in name of output file
    mineos_ref : string
        Reference file, used to copy the header and for the rest of the planet,
        in the case of a Layer(), default = 'mineos_prem_noocean.txt'
    plotting: Boolean
        True means plot of the old model and replaced model will be shown
        (default = False)

    """

    if not isinstance(layers[0], Layer):
        raise TypeError("Input must be a list of Layer()")

    # Load reference input
    datastream = pkgutil.get_data(
        'burnman', 'data/input_seismic/' + mineos_ref)
    lines = [line.strip()
             for line in datastream.decode('ascii').split('\n') if line.strip()]
    table = []
    for line in lines[3:]:
        numbers = np.fromstring(line, sep=' ')
        table.append(numbers)
    table = np.array(table)

    if plotting:
        plt.figure(figsize=(12, 6))
        plt.plot(table[:, 0] / 1.e3, table[:, 2] / 1.e3,
                 color='g', linestyle='--')
        plt.plot(table[:, 0] / 1.e3, table[:, 3] / 1.e3,
                 color='b', linestyle='--')
        plt.plot(table[:, 0] / 1.e3, table[:, 1] / 1.e3,
                 color='r', linestyle='--')

    for layer in layers:
        i_min = next(x[0] for x in enumerate(table[:, 0])
                     if x[1] >= layer.inner_radius)
        if table[i_min, 0] - layer.inner_radius > 0:
            table[i_min, 0] = layer.inner_radius

        i_max = next(x[0] for x in enumerate(table[:, 0])
                     if x[1] >= layer.outer_radius)
        if table[i_max, 0] - layer.outer_radius > 0:
            table[i_max, 0] = layer.outer_radius

        layer_vp, layer_vs, layer_density = layer.evaluate(
            ['v_p', 'v_s', 'density'], radlist=table[i_min:i_max, 0], radius_planet=np.max(table[:, 0]))

        table[i_min:i_max, 2] = layer_vp
        table[i_min:i_max, 6] = layer_vp
        table[i_min:i_max, 3] = layer_vs
        table[i_min:i_max, 7] = layer_vs
        table[i_min:i_max, 1] = layer_density

    # WRITE OUT FILE
    filename = 'mineos_' + modelname + '.txt'
    f = open(filename, 'w')
    print('Writing ' + filename + ' ...')
    f.write(lines[0][:-2] + ' +  ' + filename + '\n')
    for line in lines[1:3]:
        f.write(line[:-2] + '\n')
    for i in range(len(table[:, 0])):
        f.write('%8.0f %9.2f %9.2f %9.2f %9.1f %9.1f %9.2f %9.2f %9.5f \n' % (
            table[i, 0], table[i, 1], table[i, 2], table[i, 3], table[i, 4], table[i, 5], table[i, 6], table[i, 7], table[i, 8]))

    f.close()

    if plotting:

        plt.plot(table[:, 0] / 1.e3, table[:, 2] / 1.e3,
                 color='g', linestyle='-', label='V_p')
        plt.plot(table[:, 0] / 1.e3, table[:, 3] / 1.e3,
                 color='b', linestyle='-', label='V_s')
        plt.plot(table[:, 0] / 1.e3, table[:, 1] / 1.e3,
                 color='r', linestyle='-', label='density')

        plt.title(f'{filename} = {mineos_ref} replaced between '
                  f'{str(layer.inner_radius / 1.e3)} and '
                  f'{str(layer.outer_radius / 1.e3)} km')
        plt.legend(loc='lower right')
        plt.show()



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


import scipy.optimize as opt
from . import equation_of_state as eos
import warnings
from math import exp


def bulk_modulus(volume, params):
    """
    compute the bulk modulus as per the
    Vinet equation of state.  Reference bulk
    modulus should be in :math:`[Pa]`.
    """

    x = volume / params['V_0']
    eta = (3. / 2.) * (params['Kprime_0'] - 1.)

    K = (params['K_0'] * pow(x, -2. / 3.)) * \
        (1 + ((eta * pow(x, 1. / 3.) + 1.) * (1. - pow(x, 1. / 3.)))) * \
        exp(eta * (1. - pow(x, 1. / 3.)))
    return K


def vinet(x, params):
    """
    equation for the  Vinet equation of state, returns
    pressure in the same units that are supplied for the reference bulk
    modulus (params['K_0']), which should be in math:`[Pa]`.
    """
    eta = (3. / 2.) * (params['Kprime_0'] - 1.)
    return 3. * params['K_0'] * (pow(x, -2. / 3.)) * (1. - (pow(x, 1. / 3.))) \
        * exp(eta * (1. - pow(x, 1. / 3.))) + params['P_0']


def volume(pressure, params):
    """
    Get the Vinet volume at a reference temperature for a given
    pressure :math:`[Pa]`. Returns molar volume in :math:`[m^3]`
    """

    func = lambda x: vinet(x / params['V_0'], params) - pressure
    V = opt.brentq(func, 0.1 * params['V_0'], 1.5 * params['V_0'])
    return V


class Vinet(eos.EquationOfState):

    """
    Base class for the isothermal Vinet equation of state.
    References for this equation of state are :cite:`vinet1986`
    and :cite:`vinet1987`. This equation of state actually
    predates Vinet by 55 years :cite:`Rydberg1932`,
    and was investigated further by :cite:`Stacey1981`.
    """

    def volume(self, pressure, temperature, params):
        """
        Returns volume :math:`[m^3]` as a function of pressure :math:`[Pa]`.
        """
        return volume(pressure, params)

    def pressure(self, temperature, volume, params):
        return vinet(volume / params['V_0'], params)

    def isothermal_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`K_T` :math:`[Pa]` as a function of pressure :math:`[Pa]`,
        temperature :math:`[K]` and volume :math:`[m^3]`.
        """
        return bulk_modulus(volume, params)

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus :math:`K_s` of the mineral. :math:`[Pa]`.
        """
        return bulk_modulus(volume, params)

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus :math:`G` of the mineral. :math:`[Pa]`
        Currently not included in the Vinet EOS, so omitted.
        """
        return 0.

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the molar entropy :math:`\mathcal{S}` of the mineral. :math:`[J/K/mol]`
        """
        return 0.

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy :math:`\mathcal{E}` of the mineral. :math:`[J/mol]`
        """
        x = pow(volume/params['V_0'], 1./3.)
        eta = (3. / 2.) * (params['Kprime_0'] - 1.)


        intPdV = (9.* params['V_0'] * params['K_0'] / (eta*eta) *
                  ((1. - eta*(1. - x))*exp(eta*(1. - x)) - 1.))


        return - intPdV + params['E_0']

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy :math:`\mathcal{G}` of the mineral. :math:`[J/mol]`
        """
        # G = int VdP = [PV] - int PdV = E + PV

        return self.molar_internal_energy(pressure, temperature, volume, params) + volume*pressure

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.e99

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.e99

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[1/K]`
        """
        return 0.

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[unitless]`
        """
        return 0.

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """


        if 'E_0' not in params:
            params['E_0'] = 0.
        if 'P_0' not in params:
            params['P_0'] = 0.

        # G is not included in the Vinet EOS so we shall set them to NaN's
        if 'G_0' not in params:
            params['G_0'] = float('nan')
        if 'Gprime_0' not in params:
            params['Gprime_0'] = float('nan')

        # check that all the required keys are in the dictionary
        expected_keys = ['V_0', 'K_0', 'Kprime_0']
        for k in expected_keys:
            if k not in params:
                raise KeyError('params object missing parameter : ' + k)

        # now check that the values are reasonable.  I mostly just
        # made up these values from experience, and we are only
        # raising a warning.  Better way to do this? [IR]
        if params['V_0'] < 1.e-7 or params['V_0'] > 1.e-3:
            warnings.warn('Unusual value for V_0', stacklevel=2)
        if params['K_0'] < 1.e9 or params['K_0'] > 1.e13:
            warnings.warn('Unusual value for K_0', stacklevel=2)
        if params['Kprime_0'] < -5. or params['Kprime_0'] > 10.:
            warnings.warn('Unusual value for Kprime_0', stacklevel=2)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


from __future__ import absolute_import

import numpy as np
import warnings
import scipy.integrate

from ..tools.misc import read_table
from .. import constants


class Seismic1DModel(object):

    """
    Base class for all the seismological models.
    """

    def __init__(self):
        pass

    def evaluate(self, vars_list, depth_list=None):
        """
        Returns the lists of data for a Seismic1DModel for the depths provided

        Parameters
        ----------
        vars_list : array of str
            Available variables depend on the seismic model, and can be chosen from 'pressure','density','gravity','v_s','v_p','v_phi','G','K','QG','QK'
        depth_list : array of floats
            Array of depths [m] to evaluate seismic model at.

        Returns
        -------
        Array of values shapes as (len(vars_list),len(depth_list)).

        """
        if depth_list is None:
            depth_list = self.internal_depth_list()
        values = np.empty((len(vars_list), len(depth_list)))
        for a in range(len(vars_list)):
            values[a, :] = getattr(self, vars_list[a])(depth_list)
        return values

    def internal_depth_list(self, mindepth=0., maxdepth=1.e99, discontinuity_interval=1.):
        """
        Returns a sorted list of depths where this seismic data is specified at. This allows you to compare the seismic data without interpolation. The depths can be bounded by the mindepth and maxdepth parameters.

        Parameters
        ----------
        mindepth  :  float
            Minimum depth value to be returned [m]
        maxdepth
            Maximum depth value to be returned [m]
        discontinuity interval
            Shift continuities to remove ambigious values for depth, default value = 1 [m]

        Returns
        -------
        depths : array of floats
            Depths [m].
        """
        raise NotImplementedError(
            "abstract method to be implemented in derived class")

    def pressure(self, depth):
        """
        Parameters
        ----------
        depth : float or array of floats
            Depth(s) [m] to evaluate seismic model at.

        Returns
        -------
        pressure : float or array of floats
            Pressure(s) at given depth(s) in [Pa].
        """
        raise NotImplementedError(
            "abstract method to be implemented in derived class")

    def v_p(self, depth):
        """
        Parameters
        ----------
        depth : float or array of floats
            Depth(s) [m] to evaluate seismic model at.

        Returns
        -------
        v_p : float or array of floats
            P wave velocity at given depth(s) in [m/s].
        """
        raise NotImplementedError(
            "abstract method to be implemented in derived class")

    def v_s(self, depth):
        """
        Parameters
        ----------
        depth : float or array of floats
            Depth(s) [m] to evaluate seismic model at.

        Returns
        -------
        v_s : float or array of floats
            S wave velocity at given depth(s) in [m/s].
        """
        raise NotImplementedError(
            "abstract method to be implemented in derived class")

    def v_phi(self, depth):
        """
        Parameters
        ----------
        depth_list : float or array of floats
            Depth(s) [m] to evaluate seismic model at.

        Returns
        -------
        v_phi : float or array of floats
            bulk sound wave velocity at given depth(s) in [m/s].
        """
        v_s = self.v_s(depth)
        v_p = self.v_p(depth)
        return np.sqrt(v_p * v_p - 4. / 3. * v_s * v_s)

    def density(self, depth):
        """
        Parameters
        ----------
        depth : float or array of floats
            Depth(s) [m] to evaluate seismic model at.

        Returns
        -------
        density : float or array of floats
            Density at given depth(s) in [kg/m^3].
        """
        raise NotImplementedError(
            "abstract method to be implemented in derived class")

    def G(self, depth):
        """
        Parameters
        ----------
        depth : float or array of floats
            Shear modulus at given for depth(s) in [Pa].
        """
        return np.power(self.v_s(depth), 2.) * self.density(depth)

    def K(self, depth):
        """
        Parameters
        ----------
        depth : float or array of floats
            Bulk modulus at given for depth(s) in [Pa]
        """
        return np.power(self.v_phi(depth), 2.) * self.density(depth)

    def QK(self, depth):
        """
        Parameters
        ----------
        depth : float or array of floats
            Depth(s) [m] to evaluate seismic model at.

        Returns
        -------
        Qk : float or array of floats
            Quality factor (dimensionless) for bulk modulus at given depth(s).
        """
        raise NotImplementedError(
            "abstract method to be implemented in derived class")

    def QG(self, depth):
        """
        Parameters
        ----------
        depth : float or array of floats
            Depth(s) [m] to evaluate seismic model at.

        Returns
        -------
        QG : float or array of floats
            Quality factor (dimensionless) for shear modulus at given depth(s).
        """
        raise NotImplementedError(
            "abstract method to be implemented in derived class")

    def depth(self, pressure):
        """
        Parameters
        ----------
        pressure : float or array of floats
            Pressure(s) [Pa] to evaluate depth at.

        Returns
        -------
        depth : float or array of floats
            Depth(s) [m] for given pressure(s)
        """
        raise NotImplementedError(
            "abstract method to be implemented in derived class")

    def gravity(self, depth):
        """
        Parameters
        ----------
        depth : float or array of floats
            Depth(s) [m] to evaluate gravity at.

        Returns
        -------
        gravity : float or array of floats
            Gravity for given depths in [m/s^2]
        """
        raise NotImplementedError(
            "abstract method to be implemented in derived class")


class SeismicTable(Seismic1DModel):
    """
    This is a base class that gets a 1D seismic model from a table indexed and
    sorted by radius. Fill the tables in the constructor after deriving
    from this class. This class uses :class:`burnman.seismic.Seismic1DModel`

    Note: all tables need to be sorted by increasing depth. self.table_depth needs to be defined
    Alternatively, you can also overwrite the _lookup function if you
    want to access with something else.
    """

    def __init__(self):
        Seismic1DModel.__init__(self)

        self.table_depth = []
        self.table_radius = []
        self.table_pressure = []
        self.table_gravity = []
        self.table_density = []
        self.table_vp = []
        self.table_vs = []
        self.table_QG = []
        self.table_QK = []

        self.earth_radius = 6371.0e3

    def internal_depth_list(self, mindepth=0., maxdepth=1.e10, discontinuity_interval=1.):
        depths = np.array([self.table_depth[x] for x in range(len(
            self.table_depth)) if self.table_depth[x] >= mindepth and self.table_depth[x] <= maxdepth])
        discontinuities = np.where(depths[1:] - depths[:-1] == 0)[0]
        # Shift values at discontinities by 1 m to simplify evaluating values
        # around these.
        depths[discontinuities] = depths[discontinuities] - discontinuity_interval
        depths[discontinuities + 1] = depths[discontinuities + 1] + discontinuity_interval
        return depths

    def pressure(self, depth):
        if len(self.table_pressure) == 0:
            warnings.warn(
                "Pressure is not given in " + self.__class__.__name__ + " and is now being computed. This will only work when density is defined for the entire planet. Use at your own risk.")
            self._compute_pressure()
        return self._lookup(depth, self.table_pressure)

    def gravity(self, depth):
        if len(self.table_gravity) == 0:
            warnings.warn(
                "Gravity is not given in " + self.__class__.__name__ + " and is now being computed. This will only work when density is defined for the entire planet. Use at your own risk.")
            self._compute_gravity()
        return self._lookup(depth, self.table_gravity)

    def v_p(self, depth):

        return self._lookup(depth, self.table_vp)

    def v_s(self, depth):

        return self._lookup(depth, self.table_vs)

    def QK(self, depth):

        return self._lookup(depth, self.table_QK)

    def QG(self, depth):

        return self._lookup(depth, self.table_QG)

    def density(self, depth):
        if len(self.table_density) == 0:
            raise ValueError(
                "Density has not been defined for this seismic model")
        return self._lookup(depth, self.table_density)

    def bullen(self, depth):
        """
        Returns the Bullen parameter only for significant arrays
        """
        assert(len(depth) > 3)
        v_phi = self.v_phi(depth)
        density = self.density(depth)
        phi = v_phi * v_phi
        kappa = phi * density
        try:
            dkappadP = np.gradient(kappa, edge_order=2) / \
                       np.gradient(self.pressure(depth), edge_order=2)
            dphidz = np.gradient(phi,
                                 edge_order=2) / np.gradient(depth,
                                                             edge_order=2) / self.gravity(depth)
        except:
            dkappadP = np.gradient(kappa) / np.gradient(self.pressure(depth))
            dphidz = np.gradient(phi) / np.gradient(depth) / self.gravity(depth)
        bullen = dkappadP - dphidz
        return bullen

    def depth(self, pressure):
        if max(pressure) > max(
                self.table_pressure) or min(pressure) < min(
                self.table_pressure):
            raise ValueError("Pressure outside range of SeismicTable")

        depth = np.interp(pressure, self.table_pressure, self.table_depth)
        return depth

    def radius(self, pressure):

        radius = np.interp(pressure, self.table_pressure[
                           ::-1], self.earth_radius - self.table_depth[::-1])
        return radius

    def _lookup(self, depth, value_table):
        return np.interp(depth, self.table_depth, value_table)

    def _compute_gravity(self):
        # Calculate the gravity of the planet, based on a density profile.
        # There is a check of the surface value is within reason for this
        # model, otherwise values for PREM are used.

        density = self.table_density[::-1]

        if len(density) > 0:
            radii = self.table_radius[::-1]
            g = scipy.integrate.cumtrapz(constants.G * 4. * np.pi
                                         * density * radii * radii,
                                         x=radii,
                                         initial=0)
            g[1:] = g[1:] / radii[1:] / radii[1:]

            self.table_gravity = g[::-1]
        else:
            raise ValueError('Density profile is an empty list '
                             'in this SeismicTable instance.')

    def _compute_pressure(self):
        # Calculate the pressure profile based on density and gravity.
        # This integrates the equation for hydrostatic equilibrium P = rho g z.
        radii = self.table_radius
        density = self.table_density
        gravity = self.gravity(self.earth_radius - radii)

        # convert radii to depths
        depth = self.earth_radius - radii
        pressure = scipy.integrate.cumtrapz(
            gravity * density, x=depth, initial=0.)

        self.table_pressure = pressure


class PREM(SeismicTable):

    """
    Reads  PREM (1s) (input_seismic/prem.txt, :cite:`dziewonski1981`).
    See also :class:`burnman.seismic.SeismicTable`.
    """

    def __init__(self):
        SeismicTable.__init__(self)
        table = read_table("input_seismic/prem.txt")
        table = np.array(table)
        self.table_depth = table[:, 0]
        self.table_radius = table[:, 1]
        self.table_pressure = table[:, 2]
        self.table_density = table[:, 3]
        self.table_vp = table[:, 4]
        self.table_vs = table[:, 5]
        self.table_QK = table[:, 6]
        self.table_QG = table[:, 7]


class Slow(SeismicTable):

    """
    Inserts the mean profiles for slower regions in the lower mantle (Lekic et al. 2012).
    We stitch together tables 'input_seismic/prem_lowermantle.txt', 'input_seismic/swave_slow.txt', 'input_seismic/pwave_slow.txt').
    See also :class:`burnman.seismic.SeismicTable`.
    """

    def __init__(self):
        SeismicTable.__init__(self)

        # data is: depth radius pressure density V_p V_s Q_K Q_G
        table = read_table("input_seismic/prem.txt")
        table = np.array(table)
        table2 = read_table("input_seismic/swave_slow.txt")
        table2 = np.array(table2)
        table3 = read_table("input_seismic/pwave_slow.txt")
        table3 = np.array(table3)

        min_radius = self.earth_radius - max(table2[:, 0])
        max_radius = self.earth_radius - min(table2[:, 0])

        table = np.array(
            list(
                filter(
                    lambda x: (
                        x[1] >= min_radius and x[1] <= max_radius),
                    table)))

        self.table_depth = table[:, 0]
        self.table_radius = table[:, 1]
        self.table_pressure = table[:, 2]
        self.table_density = table[:, 3]
        self.table_vp = np.interp(
            self.table_depth, table3[:, 0][::-1], table3[:, 1][::-1])
        self.table_vs = np.interp(
            self.table_depth, table2[:, 0][::-1], table2[:, 1][::-1])


class Fast(SeismicTable):

    """
    Inserts the mean profiles for faster regions in the lower mantle (Lekic et al. 2012).
    We stitch together tables 'input_seismic/prem_lowermantle.txt', 'input_seismic/swave_fast.txt', 'input_seismic/pwave_fast.txt').
    See also :class:`burnman.seismic.Seismic1DModel`.
    """

    def __init__(self):
        SeismicTable.__init__(self)

        # data is: radius pressure density V_p V_s Q_K Q_G
        table = read_table("input_seismic/prem.txt")
        table = np.array(table)
        table2 = read_table("input_seismic/swave_fast.txt")
        table2 = np.array(table2)
        table3 = read_table("input_seismic/pwave_fast.txt")
        table3 = np.array(table3)

        min_radius = self.earth_radius - max(table2[:, 0])
        max_radius = self.earth_radius - min(table2[:, 0])

        table = np.array(
            list(
                filter(
                    lambda x: (
                        x[1] >= min_radius and x[1] <= max_radius),
                    table)))

        self.table_depth = table[:, 0]
        self.table_radius = table[:, 1]
        self.table_pressure = table[:, 2]
        self.table_density = table[:, 3]
        self.table_vp = np.interp(
            self.table_depth, table3[:, 0][::-1], table3[:, 1][::-1])
        self.table_vs = np.interp(
            self.table_depth, table2[:, 0][::-1], table2[:, 1][::-1])


class STW105(SeismicTable):

    """
    Reads  STW05 (a.k.a. REF) (1s) (input_seismic/STW105.txt, :cite:`kustowski2008`).
    See also :class:`burnman.seismic.SeismicTable`.
    """

    def __init__(self):
        SeismicTable.__init__(self)
        # radius, pressure, density, v_p, v_s
        table = read_table("input_seismic/STW105.txt")
        table = np.array(table)
        self.table_radius = table[:, 0][::-1]
        self.table_density = table[:, 1][::-1]
        self.table_vpv = table[:, 2][::-1]
        self.table_vsv = table[:, 3][::-1]
        self.table_QK = table[:, 4][::-1]
        self.table_QG = table[:, 5][::-1]
        self.table_vph = table[:, 6][::-1]
        self.table_vsh = table[:, 7][::-1]

        self.table_depth = self.earth_radius - self.table_radius

        # Voigt averages for Vs and Vp
        self.table_vs = np.sqrt((2. * self.table_vsv * self.table_vsv + self.table_vsh * self.table_vsh) / 3.)
        self.table_vp = np.sqrt((self.table_vpv * self.table_vpv + 4. * self.table_vph * self.table_vph) / 5.)


class IASP91(SeismicTable):
    """
    Reads  REF/STW05 (input_seismic/STW105.txt, :cite:`kustowski2008`).
    See also :class:`burnman.seismic.SeismicTable`.
    """

    def __init__(self):
        SeismicTable.__init__(self)
        table = read_table(
            "input_seismic/iasp91.txt")  # depth, radius, v_p, v_s
        table = np.array(table)
        self.table_depth = table[:, 0]
        self.table_radius = table[:, 1]
        self.table_vp = table[:, 2]
        self.table_vs = table[:, 3]


class AK135(SeismicTable):
    """
    Reads  AK135 (input_seismic/ak135.txt, :cite:`kennett1995`).
    See also :class:`burnman.seismic.SeismicTable`.
    """

    def __init__(self):
        SeismicTable.__init__(self)
        table = read_table(
            "input_seismic/ak135.txt")  # radius, pressure, density, v_p, v_s
        table = np.array(table)
        self.table_depth = table[:, 0]
        self.table_radius = table[:, 1]
        self.table_density = table[:, 2]
        self.table_vp = table[:, 3]
        self.table_vs = table[:, 4]
        self.table_QG = table[:, 5]
        self.table_QK = table[:, 6]


def attenuation_correction(v_p, v_s, v_phi, Qs, Qphi):
    """
    Applies the attenuation correction following Matas et al. (2007), page 4. This is simplified, and there is also currently no 1D Q model implemented. The correction, however, only slightly reduces the velocities, and can be ignored for our current applications. Arguably, it might not be as relevant when comparing computations to PREM for periods of 1s as is implemented here.
    Called from :func:`burnman.main.apply_attenuation_correction`

    Parameters
    ----------
    v_p : float
        P wave velocity in [m/s].
    v_s : float
        S wave velocitiy in [m/s].
    v_phi : float
        Bulk sound velocity in [m/s].
    Qs : float
        shear quality factor [dimensionless]
    Qphi: float
        bulk quality factor [dimensionless]


    Returns
    -------
    v_p : float
        corrected P wave velocity in [m/s].
    v_s : float
        corrected S wave velocitiy in [m/s].
    v_phi : float
        corrected Bulk sound velocity in [m/s].
    """
    beta = 0.3  # Matas et al. (2007) page 4
    Qp = 3. / 4. * pow((v_p / v_s), 2.) * Qs    # Matas et al. (2007) page 4

    cot = 1. / np.tan(beta * np.pi / 2.)
    v_p *= 1. - 1. / 2. * cot * 1. / Qp  # Matas et al. (2007) page 1
    v_s *= 1. - 1. / 2. * cot * 1. / Qs
    v_phi *= 1. - 1. / 2. * cot * 1. / Qphi
    return v_p, v_s, v_phi


"""
shared variable of prem, so that other routines do not need to create
prem over and over. See geotherm for example.
"""
prem_model = PREM()



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.


from __future__ import absolute_import

import numpy as np
from sympy import Matrix, nsimplify
from .material import material_property, cached_property
from .mineral import Mineral
from .solutionmodel import SolutionModel
from .solutionmodel import MechanicalSolution, IdealSolution
from .solutionmodel import SymmetricRegularSolution, AsymmetricRegularSolution
from .solutionmodel import SubregularSolution
from .averaging_schemes import reuss_average_function

from ..tools.reductions import independent_row_indices
from ..tools.chemistry import sum_formulae, sort_element_list_to_IUPAC_order


class SolidSolution(Mineral):
    """
    This is the base class for all solid solutions.
    Site occupancies, endmember activities and the constant
    and pressure and temperature dependencies of the excess
    properties can be queried after using set_composition()
    States of the solid solution can only be queried after setting
    the pressure, temperature and composition using set_state().

    This class is available as :class:`burnman.SolidSolution`.
    It uses an instance of :class:`burnman.SolutionModel` to
    calculate interaction terms between endmembers.

    All the solid solution parameters are expected to be in SI units.  This
    means that the interaction parameters should be in J/mol, with the T
    and P derivatives in J/K/mol and m^3/mol.

    The parameters are relevant to all solution models. Please
    see the documentation for individual models for details about
    other parameters.

    Parameters
    ----------
    name : string
        Name of the solid solution
    solution_type : string
        String determining which SolutionModel to use. One of 'mechanical',
        'ideal', 'symmetric', 'asymmetric' or 'subregular'.
    endmembers : list of lists
        List of endmembers in this solid solution. The first item of each
        list should be a :class:`burnman.Mineral` object. The second item
        should be a string with the site formula of the endmember.
    molar_fractions : numpy array (optional)
        The molar fractions of each endmember in the solid solution.
        Can be reset using the set_composition() method.
    """

    def __init__(self,
                 name=None,
                 solution_type=None,
                 endmembers=None,
                 energy_interaction=None,
                 volume_interaction=None,
                 entropy_interaction=None,
                 energy_ternary_terms=None,
                 volume_ternary_terms=None,
                 entropy_ternary_terms=None,
                 alphas=None,
                 molar_fractions=None):
        """
        Set up matrices to speed up calculations for when P, T, X is defined.
        """
        Mineral.__init__(self)

        # SolidSolution needs a method attribute to call Mineral.set_state().
        # Note that set_method() below will not change self.method
        self.method = 'SolidSolutionMethod'

        if name is not None:
            self.name = name
        if solution_type is not None:
            self.solution_type = solution_type
        if endmembers is not None:
            self.endmembers = endmembers
        if energy_interaction is not None:
            self.energy_interaction = energy_interaction
        if volume_interaction is not None:
            self.volume_interaction = volume_interaction
        if entropy_interaction is not None:
            self.entropy_interaction = entropy_interaction
        if energy_ternary_terms is not None:
            self.energy_ternary_terms = energy_ternary_terms
        if volume_ternary_terms is not None:
            self.volume_ternary_terms = volume_ternary_terms
        if entropy_ternary_terms is not None:
            self.entropy_ternary_terms = entropy_ternary_terms
        if alphas is not None:
            self.alphas = alphas
        if endmembers is not None:
            self.endmembers = endmembers

        if hasattr(self, 'endmembers') is False:
            raise Exception("'endmembers' attribute missing "
                            "from solid solution")

        # Set default solution model type
        if hasattr(self, 'solution_type'):
            if self.solution_type == 'mechanical':
                self.solution_model = MechanicalSolution(self.endmembers)
            elif self.solution_type == 'ideal':
                self.solution_model = IdealSolution(self.endmembers)
            else:
                if hasattr(self, 'energy_interaction') is False:
                    self.energy_interaction = None
                if hasattr(self, 'volume_interaction') is False:
                    self.volume_interaction = None
                if hasattr(self, 'entropy_interaction') is False:
                    self.entropy_interaction = None

                if self.solution_type == 'symmetric':
                    self.solution_model = SymmetricRegularSolution(
                        self.endmembers, self.energy_interaction,
                        self.volume_interaction, self.entropy_interaction)
                elif self.solution_type == 'asymmetric':
                    if hasattr(self, 'alphas') is False:
                        raise Exception(
                            "'alphas' attribute missing from solid solution")
                    self.solution_model = AsymmetricRegularSolution(
                        self.endmembers, self.alphas, self.energy_interaction,
                        self.volume_interaction, self.entropy_interaction)
                elif self.solution_type == 'subregular':
                    if hasattr(self, 'energy_ternary_terms') is False:
                        self.energy_ternary_terms = None
                    if hasattr(self, 'volume_ternary_terms') is False:
                        self.volume_ternary_terms = None
                    if hasattr(self, 'entropy_ternary_terms') is False:
                        self.entropy_ternary_terms = None

                    self.solution_model = SubregularSolution(
                        self.endmembers,
                        self.energy_interaction,  self.volume_interaction,
                        self.entropy_interaction,
                        self.energy_ternary_terms, self.volume_ternary_terms,
                        self.entropy_ternary_terms)
                else:
                    raise Exception("Solution model type "
                                    + self.solution_type + "not recognised.")
        else:
            self.solution_model = SolutionModel()

        # Equation of state
        for i in range(self.n_endmembers):
            self.endmembers[i][0].set_method(
                self.endmembers[i][0].params['equation_of_state'])

        # Molar fractions
        if molar_fractions is not None:
            self.set_composition(molar_fractions)

    def get_endmembers(self):
        return self.endmembers

    def set_composition(self, molar_fractions):
        """
        Set the composition for this solid solution.
        Resets cached properties.

        Parameters
        ----------
        molar_fractions: list of float
            molar abundance for each endmember, needs to sum to one.
        """
        assert(len(self.endmembers) == len(molar_fractions))

        if self.solution_type != 'mechanical':
            assert(sum(molar_fractions) > 0.9999)
            assert(sum(molar_fractions) < 1.0001)

        self.reset()
        self.molar_fractions = np.array(molar_fractions)

    def set_method(self, method):
        for i in range(self.n_endmembers):
            self.endmembers[i][0].set_method(method)
        # note: do not set self.method here!
        self.reset()

    def set_state(self, pressure, temperature):

        Mineral.set_state(self, pressure, temperature)
        for i in range(self.n_endmembers):
            self.endmembers[i][0].set_state(pressure, temperature)

    @material_property
    def formula(self):
        """
        Returns molar chemical formula of the solid solution.
        """
        return sum_formulae(self.endmember_formulae, self.molar_fractions)

    @material_property
    def activities(self):
        """
        Returns a list of endmember activities [unitless].
        """
        return self.solution_model.activities(self.pressure, self.temperature,
                                              self.molar_fractions)

    @material_property
    def activity_coefficients(self):
        """
        Returns a list of endmember activity coefficients
        (gamma = activity / ideal activity) [unitless].
        """
        return self.solution_model.activity_coefficients(self.pressure,
                                                         self.temperature,
                                                         self.molar_fractions)

    @material_property
    def molar_internal_energy(self):
        """
        Returns molar internal energy of the mineral [J/mol].
        Aliased with self.energy
        """
        return self.molar_helmholtz + self.temperature * self.molar_entropy

    @material_property
    def excess_partial_gibbs(self):
        """
        Returns excess partial molar gibbs free energy [J/mol].
        Property specific to solid solutions.
        """
        return self.solution_model.excess_partial_gibbs_free_energies(self.pressure, self.temperature, self.molar_fractions)

    @material_property
    def excess_partial_volumes(self):
        """
        Returns excess partial volumes [m^3].
        Property specific to solid solutions.
        """
        return self.solution_model.excess_partial_volumes(self.pressure,
                                                          self.temperature,
                                                          self.molar_fractions)

    @material_property
    def excess_partial_entropies(self):
        """
        Returns excess partial entropies [J/K].
        Property specific to solid solutions.
        """
        return self.solution_model.excess_partial_entropies(self.pressure,
                                                            self.temperature,
                                                            self.molar_fractions)

    @material_property
    def partial_gibbs(self):
        """
        Returns excess partial molar gibbs free energy [J/mol].
        Property specific to solid solutions.
        """
        return (np.array([self.endmembers[i][0].gibbs
                          for i in range(self.n_endmembers)])
                + self.excess_partial_gibbs)

    @material_property
    def partial_volumes(self):
        """
        Returns excess partial volumes [m^3].
        Property specific to solid solutions.
        """
        return (np.array([self.endmembers[i][0].molar_volume
                          for i in range(self.n_endmembers)])
                + self.excess_partial_volumes)

    @material_property
    def partial_entropies(self):
        """
        Returns excess partial entropies [J/K].
        Property specific to solid solutions.
        """
        return (np.array([self.endmembers[i][0].molar_entropy
                          for i in range(self.n_endmembers)])
                + self.excess_partial_entropies)

    @material_property
    def excess_gibbs(self):
        """
        Returns molar excess gibbs free energy [J/mol].
        Property specific to solid solutions.
        """
        return self.solution_model.excess_gibbs_free_energy(self.pressure,
                                                            self.temperature,
                                                            self.molar_fractions)

    @material_property
    def gibbs_hessian(self):
        """
        Returns an array containing the second compositional derivative
        of the Gibbs free energy [J]. Property specific to solid solutions.
        """
        return self.solution_model.gibbs_hessian(self.pressure,
                                                 self.temperature,
                                                 self.molar_fractions)

    @material_property
    def entropy_hessian(self):
        """
        Returns an array containing the second compositional derivative
        of the entropy [J/K]. Property specific to solid solutions.
        """
        return self.solution_model.entropy_hessian(self.pressure,
                                                   self.temperature,
                                                   self.molar_fractions)

    @material_property
    def volume_hessian(self):
        """
        Returns an array containing the second compositional derivative
        of the volume [m^3]. Property specific to solid solutions.
        """
        return self.solution_model.volume_hessian(self.pressure,
                                                  self.temperature,
                                                  self.molar_fractions)

    @material_property
    def molar_gibbs(self):
        """
        Returns molar Gibbs free energy of the solid solution [J/mol].
        Aliased with self.gibbs.
        """
        return sum([self.endmembers[i][0].gibbs * self.molar_fractions[i]
                    for i in range(self.n_endmembers)]) + self.excess_gibbs

    @material_property
    def molar_helmholtz(self):
        """
        Returns molar Helmholtz free energy of the solid solution [J/mol].
        Aliased with self.helmholtz.
        """
        return self.molar_gibbs - self.pressure * self.molar_volume

    @material_property
    def molar_mass(self):
        """
        Returns molar mass of the solid solution [kg/mol].
        """
        return sum([self.endmembers[i][0].molar_mass
                    * self.molar_fractions[i]
                    for i in range(self.n_endmembers)])

    @material_property
    def excess_volume(self):
        """
        Returns excess molar volume of the solid solution [m^3/mol].
        Specific property for solid solutions.
        """
        return self.solution_model.excess_volume(self.pressure,
                                                 self.temperature,
                                                 self.molar_fractions)

    @material_property
    def molar_volume(self):
        """
        Returns molar volume of the solid solution [m^3/mol].
        Aliased with self.V.
        """
        return sum([self.endmembers[i][0].molar_volume
                    * self.molar_fractions[i]
                    for i in range(self.n_endmembers)]) + self.excess_volume

    @material_property
    def density(self):
        """
        Returns density of the solid solution [kg/m^3].
        Aliased with self.rho.
        """
        return self.molar_mass / self.molar_volume

    @material_property
    def excess_entropy(self):
        """
        Returns excess molar entropy [J/K/mol].
        Property specific to solid solutions.
        """
        return self.solution_model.excess_entropy(self.pressure,
                                                  self.temperature,
                                                  self.molar_fractions)

    @material_property
    def molar_entropy(self):
        """
        Returns molar entropy of the solid solution [J/K/mol].
        Aliased with self.S.
        """
        return sum([self.endmembers[i][0].S * self.molar_fractions[i]
                    for i in range(self.n_endmembers)]) + self.excess_entropy

    @material_property
    def excess_enthalpy(self):
        """
        Returns excess molar enthalpy [J/mol].
        Property specific to solid solutions.
        """
        return self.solution_model.excess_enthalpy(self.pressure,
                                                   self.temperature,
                                                   self.molar_fractions)

    @material_property
    def molar_enthalpy(self):
        """
        Returns molar enthalpy of the solid solution [J/mol].
        Aliased with self.H.
        """
        return sum([self.endmembers[i][0].H * self.molar_fractions[i]
                    for i in range(self.n_endmembers)]) + self.excess_enthalpy

    @material_property
    def isothermal_bulk_modulus(self):
        """
        Returns isothermal bulk modulus of the solid solution [Pa].
        Aliased with self.K_T.
        """
        return self.V * 1. / (sum([self.endmembers[i][0].V
                                   / (self.endmembers[i][0].K_T)
                                   * self.molar_fractions[i]
                                   for i in range(self.n_endmembers)]))

    @material_property
    def adiabatic_bulk_modulus(self):
        """
        Returns adiabatic bulk modulus of the solid solution [Pa].
        Aliased with self.K_S.
        """
        if self.temperature < 1e-10:
            return self.isothermal_bulk_modulus
        else:
            return (self.isothermal_bulk_modulus
                    * self.molar_heat_capacity_p / self.molar_heat_capacity_v)

    @material_property
    def isothermal_compressibility(self):
        """
        Returns isothermal compressibility of the solid solution.
        (or inverse isothermal bulk modulus) [1/Pa].
        Aliased with self.K_T.
        """
        return 1. / self.isothermal_bulk_modulus

    @material_property
    def adiabatic_compressibility(self):
        """
        Returns adiabatic compressibility of the solid solution.
        (or inverse adiabatic bulk modulus) [1/Pa].
        Aliased with self.K_S.
        """
        return 1. / self.adiabatic_bulk_modulus

    @material_property
    def shear_modulus(self):
        """
        Returns shear modulus of the solid solution [Pa].
        Aliased with self.G.
        """
        G_list = np.fromiter((e[0].G for e in self.endmembers), dtype=float,
                             count=self.n_endmembers)
        return reuss_average_function(self.molar_fractions, G_list)

    @material_property
    def p_wave_velocity(self):
        """
        Returns P wave speed of the solid solution [m/s].
        Aliased with self.v_p.
        """
        return np.sqrt((self.adiabatic_bulk_modulus
                        + 4. / 3. * self.shear_modulus) / self.density)

    @material_property
    def bulk_sound_velocity(self):
        """
        Returns bulk sound speed of the solid solution [m/s].
        Aliased with self.v_phi.
        """
        return np.sqrt(self.adiabatic_bulk_modulus / self.density)

    @material_property
    def shear_wave_velocity(self):
        """
        Returns shear wave speed of the solid solution [m/s].
        Aliased with self.v_s.
        """
        return np.sqrt(self.shear_modulus / self.density)

    @material_property
    def grueneisen_parameter(self):
        """
        Returns grueneisen parameter of the solid solution [unitless].
        Aliased with self.gr.
        """
        if self.temperature < 1e-10:
            return float('nan')
        else:
            return (self.thermal_expansivity * self.isothermal_bulk_modulus
                    * self.molar_volume / self.molar_heat_capacity_v)

    @material_property
    def thermal_expansivity(self):
        """
        Returns thermal expansion coefficient (alpha)
        of the solid solution [1/K].
        Aliased with self.alpha.
        """
        return (1. / self.V) * sum([self.endmembers[i][0].alpha
                                    * self.endmembers[i][0].V
                                    * self.molar_fractions[i]
                                    for i in range(self.n_endmembers)])

    @material_property
    def molar_heat_capacity_v(self):
        """
        Returns molar heat capacity at constant volume of the
        solid solution [J/K/mol].
        Aliased with self.C_v.
        """
        return (self.molar_heat_capacity_p
                - self.molar_volume * self.temperature
                * self.thermal_expansivity * self.thermal_expansivity
                * self.isothermal_bulk_modulus)

    @material_property
    def molar_heat_capacity_p(self):
        """
        Returns molar heat capacity at constant pressure
        of the solid solution [J/K/mol].
        Aliased with self.C_p.
        """
        return sum([self.endmembers[i][0].molar_heat_capacity_p
                    * self.molar_fractions[i]
                    for i in range(self.n_endmembers)])

    @cached_property
    def stoichiometric_matrix(self):
        """
        A sympy Matrix where each element M[i,j] corresponds
        to the number of atoms of element[j] in endmember[i].
        """
        def f(i, j):
            e = self.elements[j]
            if e in self.endmember_formulae[i]:
                return nsimplify(self.endmember_formulae[i][e])
            else:
                return 0
        return Matrix(len(self.endmember_formulae), len(self.elements), f)

    @cached_property
    def stoichiometric_array(self):
        """
        An array where each element arr[i,j] corresponds
        to the number of atoms of element[j] in endmember[i].
        """
        return np.array(self.stoichiometric_matrix)

    @cached_property
    def reaction_basis(self):
        """
        An array where each element arr[i,j] corresponds
        to the number of moles of endmember[j] involved in reaction[i].
        """
        reaction_basis = np.array([v[:] for v in
                                   self.stoichiometric_matrix.T.nullspace()])

        if len(reaction_basis) == 0:
            reaction_basis = np.empty((0, len(self.endmember_names)))

        return reaction_basis

    @cached_property
    def n_reactions(self):
        """
        The number of reactions in reaction_basis.
        """
        return len(self.reaction_basis[:, 0])

    @cached_property
    def independent_element_indices(self):
        """
        A list of an independent set of element indices. If the amounts of
        these elements are known (element_amounts),
        the amounts of the other elements can be inferred by
        -compositional_null_basis[independent_element_indices].dot(element_amounts).
        """
        return sorted(independent_row_indices(self.stoichiometric_matrix.T))

    @cached_property
    def dependent_element_indices(self):
        """
        The element indices not included in the independent list.
        """
        return [i for i in range(len(self.elements))
                if i not in self.independent_element_indices]

    @cached_property
    def compositional_null_basis(self):
        """
        An array N such that N.b = 0 for all bulk compositions that can
        be produced with a linear sum of the endmembers in the solid solution.
        """
        null_basis = np.array([v[:] for v in
                               self.stoichiometric_matrix.nullspace()])

        M = null_basis[:, self.dependent_element_indices]
        assert (M.shape[0] == M.shape[1]) and (M == np.eye(M.shape[0])).all()

        return null_basis

    @cached_property
    def endmember_formulae(self):
        """
        A list of formulae for all the endmember in the solid solution.
        """
        return [mbr[0].params['formula'] for mbr in self.endmembers]

    @cached_property
    def endmember_names(self):
        """
        A list of names for all the endmember in the solid solution.
        """
        return [mbr[0].name for mbr in self.endmembers]

    @cached_property
    def n_endmembers(self):
        """
        The number of endmembers in the solid solution.
        """
        return len(self.endmembers)

    @cached_property
    def elements(self):
        """
        A list of the elements which could be contained in the solid solution,
        returned in the IUPAC element order.
        """
        keys = []
        for f in self.endmember_formulae:
            keys.extend(f.keys())

        return sort_element_list_to_IUPAC_order(set(keys))



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
import numpy as np
from scipy.optimize import brentq
from ..tools.misc import read_table, lookup_and_interpolate
from ..tools.math import bracket


def brown_shankland(depths):
    """
    Geotherm from :cite:`Brown1981`. NOTE: Valid only above 270 km

    Parameters
    ----------
    depths : list of floats
        The list of depths at which to evaluate the geotherm. :math:`[m]`

    Returns
    -------
    temperature : list of floats
        The list of temperatures for each of the pressures. :math:`[K]`
    """

    assert(min(depths) >= min(table_brown_depth))
    assert(max(depths) <= max(table_brown_depth))
    temperature = np.empty_like(depths)
    for i, depth in enumerate(depths):
        temperature[i] = lookup_and_interpolate(
            table_brown_depth, table_brown_temperature, depth)
    return temperature


def anderson(depths):
    """
    Geotherm from :cite:`anderson1982earth`.

    Parameters
    ----------
    depths : list of floats
        The list of depths at which to evaluate the geotherm. :math:`[m]`

    Returns
    -------
    temperature : list of floats
        The list of temperatures for each of the pressures. :math:`[K]`
    """
    assert(min(depths) >= min(table_anderson_depth))
    assert(max(depths) <= max(table_anderson_depth))
    temperature = np.empty_like(depths)
    for i, depth in enumerate(depths):
        temperature[i] = lookup_and_interpolate(
                table_anderson_depth, table_anderson_temperature, depth)
    return temperature


def adiabatic(pressures, T0, rock):
    """
    This calculates a geotherm based on an anchor temperature and a rock,
    assuming that the rock's temperature follows an adiabatic gradient with
    pressure. A good first guess is provided by integrating:

    .. math::
        \\frac{\partial T}{\partial P} = \\frac{ \\gamma  T}{ K_s }

    where :math:`\\gamma` is the Grueneisen parameter and :math:`K_s` is
    the adiabatic bulk modulus.

    Parameters
    ----------

    pressures : list of floats
        The list of pressures in :math:`[Pa]` at which
        to evaluate the geotherm.

    T0 : float
        An anchor temperature, corresponding to the temperature of the first
        pressure in the list. :math:`[K]`

    rock : :class:`burnman.composite`
        Material for which we compute the adiabat.  From this material we
        must compute average Grueneisen parameters and adiabatic bulk moduli
        for each pressure/temperature.

    Returns
    -------

    temperature: list of floats
        The list of temperatures for each pressure. :math:`[K]`
    """

    rock.set_state(pressures[0], T0)
    S0 = rock.S

    delta_S = lambda T, P, rock, S0: S0 - rock.evaluate(['S'], [P], [T])[0]

    temperatures = np.empty_like(pressures)
    temperatures[0] = T0
    for i in range(1, len(pressures)):
        args = (pressures[i], rock, S0)
        sol = bracket(fn=delta_S,
                      x0=(temperatures[i-1]
                          + (rock.gr*temperatures[i-1]/rock.K_S)
                          * (pressures[i] - pressures[i-1])),
                      dx=1., args=args)
        temperatures[i] = brentq(delta_S, sol[0], sol[1], args=args)

    return temperatures


table_brown = read_table("input_geotherm/brown_81.txt")
table_brown_depth = np.array(table_brown)[:, 0]
table_brown_temperature = np.array(table_brown)[:, 1]

table_anderson = read_table("input_geotherm/anderson_82.txt")
table_anderson_depth = np.array(table_anderson)[:, 0]
table_anderson_temperature = np.array(table_anderson)[:, 1]



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
from __future__ import print_function
import numpy as np
from sympy import Matrix, nsimplify
import warnings

from .material import Material, material_property, cached_property
from .mineral import Mineral
from .solidsolution import SolidSolution
from . import averaging_schemes

from ..tools.reductions import independent_row_indices
from ..tools.chemistry import sum_formulae, sort_element_list_to_IUPAC_order


def check_pairs(phases, fractions):
    if len(fractions) < 1:
        raise Exception('ERROR: we need at least one phase')

    if len(phases) != len(fractions):
        raise Exception(
            'ERROR: different array lengths for phases and fractions')

    total = sum(fractions)
    if abs(total - 1.0) > 1e-10:
        raise Exception(
            'ERROR: list of molar fractions does not add up to one')
    for p in phases:
        if not isinstance(p, Mineral):
            raise Exception(
                'ERROR: object of type ''%s'' is not of type Mineral' % (type(p)))


# static composite of minerals/composites
class Composite(Material):

    """
    Base class for a composite material.
    The static phases can be minerals or materials,
    meaning composite can be nested arbitrarily.

    The fractions of the phases can be input
    as either 'molar' or 'mass' during instantiation,
    and modified (or initialised) after this point by
    using set_fractions.

    This class is available as ``burnman.Composite``.
    """

    def __init__(self, phases, fractions=None, fraction_type='molar', name='Unnamed composite'):
        """
        Create a composite using a list of phases and their fractions (adding to 1.0).

        Parameters
        ----------
        phases: list of :class:`burnman.Material`
            list of phases.
        fractions: list of floats
            molar or mass fraction for each phase.
        fraction_type: 'molar' or 'mass' (optional, 'molar' as standard)
            specify whether molar or mass fractions are specified.
        """

        Material.__init__(self)

        assert(len(phases) > 0)
        self.phases = phases

        if fractions is not None:
            self.set_fractions(fractions, fraction_type)
        else:
            self.molar_fractions = None

        self.set_averaging_scheme('VoigtReussHill')
        self.name = name
        self.equilibrium_tolerance = 1.e-3  # J/reaction
        self.print_precision = 4  # number of significant figures used by self.__str__

    def __str__(self):
        string='Composite: {0}'.format(self.name)
        try:
            string += '\n  P, T: {0:.{sf}g} Pa, {1:.{sf}g} K'.format(self.pressure,
                                                                     self.temperature,
                                                                     sf=self.print_precision)
        except:
            pass
        string+='\nPhase and endmember fractions:'
        for phase, fraction in zip(*self.unroll()):
            string+='\n  {0}: {1:0.{sf}f}'.format(phase.name, fraction, sf=self.print_precision)
            if isinstance(phase, SolidSolution):
               for i in range(phase.n_endmembers):
                   string+='\n    {0}: {1:0.{sf}f}'.format(phase.endmember_names[i],
                                                           phase.molar_fractions[i],
                                                           sf=self.print_precision)
        return string

    def set_fractions(self, fractions, fraction_type='molar'):
        """
        Change the fractions of the phases of this Composite.
        Resets cached properties

        Parameters
        ----------
        fractions: list or numpy array of floats
            molar or mass fraction for each phase.
        fraction_type: 'molar' or 'mass'
            specify whether molar or mass fractions are specified.
        """
        assert(len(self.phases) == len(fractions))

        if isinstance(fractions, list):
            fractions = np.array(fractions)

        try:
            total = sum(fractions)
        except TypeError:
            raise Exception(
                "Since v0.8, burnman.Composite takes an array of Materials, then an array of fractions")

        assert np.all(fractions >= -1e-12)

        self.reset()

        if abs(total - 1.0) > 1e-12:
            warnings.warn('Warning: list of fractions does not add '
                          f'up to one but {total:g}. Normalizing.')
            fractions /= total

        if fraction_type == 'molar':
            molar_fractions = fractions
        elif fraction_type == 'mass':
            molar_fractions = self._mass_to_molar_fractions(self.phases,
                                                            fractions)
        else:
            raise Exception("Fraction type not recognised. "
                            "Please use 'molar' or mass")

        # Set minimum value of a molar fraction at 0.0 (rather than -1.e-12)
        self.molar_fractions = molar_fractions.clip(0.)

    def set_method(self, method):
        """
        set the same equation of state method for all the phases in the composite
        """
        for phase in self.phases:
            phase.set_method(method)
        # Clear the cache on resetting method
        self.reset()

    def set_averaging_scheme(self, averaging_scheme):
        """
        Set the averaging scheme for the moduli in the composite.
        Default is set to VoigtReussHill, when Composite is initialized.
        """

        if type(averaging_scheme) == str:
            self.averaging_scheme = getattr(
                averaging_schemes, averaging_scheme)()
        else:
            self.averaging_scheme = averaging_scheme
        # Clear the cache on resetting averaging scheme
        self.reset()

    def set_state(self, pressure, temperature):
        """
        Update the material to the given pressure [Pa] and temperature [K].
        """
        Material.set_state(self, pressure, temperature)
        for phase in self.phases:
            phase.set_state(pressure, temperature)

    def debug_print(self, indent=""):
        print("{0}Composite: {1}".format(indent, self.name))
        indent += "  "
        if self.molar_fractions is None:
            for i, phase in enumerate(self.phases):
                phase.debug_print(indent + "  ")
        else:
            for i, phase in enumerate(self.phases):
                print("%s%g of" % (indent, self.molar_fractions[i]))
                phase.debug_print(indent + "  ")

    def unroll(self):
        if self.molar_fractions is None:
            raise Exception(
                "Unroll only works if the composite has defined fractions.")
        phases = []
        fractions = []
        for i, phase in enumerate(self.phases):
            p_mineral, p_fraction = phase.unroll()
            check_pairs(p_mineral, p_fraction)
            fractions.extend([f * self.molar_fractions[i] for f in p_fraction])
            phases.extend(p_mineral)
        return phases, fractions

    def to_string(self):
        """
        return the name of the composite
        """
        return "{0}: {1}".format(self.__class__.__name__, self.name)

    @material_property
    def formula(self):
        """
        Returns molar chemical formula of the composite
        """
        return sum_formulae([ph.formula for ph in self.phases], self.molar_fractions)

    @material_property
    def molar_internal_energy(self):
        """
        Returns molar internal energy of the mineral [J/mol]
        Aliased with self.energy
        """
        U = sum(phase.molar_internal_energy * molar_fraction for (
                phase, molar_fraction) in zip(self.phases, self.molar_fractions))
        return U

    @material_property
    def molar_gibbs(self):
        """
        Returns molar Gibbs free energy of the composite [J/mol]
        Aliased with self.gibbs
        """
        G = sum(phase.molar_gibbs * molar_fraction for (phase, molar_fraction)
                in zip(self.phases, self.molar_fractions))
        return G

    @material_property
    def molar_helmholtz(self):
        """
        Returns molar Helmholtz free energy of the mineral [J/mol]
        Aliased with self.helmholtz
        """
        F = sum(phase.molar_helmholtz * molar_fraction for (
                phase, molar_fraction) in zip(self.phases, self.molar_fractions))
        return F

    @material_property
    def molar_volume(self):
        """
        Returns molar volume of the composite [m^3/mol]
        Aliased with self.V
        """
        volumes = np.array(
            [phase.molar_volume * molar_fraction for (phase, molar_fraction) in zip(self.phases, self.molar_fractions)])
        return np.sum(volumes)

    @material_property
    def molar_mass(self):
        """
        Returns molar mass of the composite [kg/mol]
        """
        return sum([phase.molar_mass * molar_fraction for (phase, molar_fraction) in zip(self.phases, self.molar_fractions)])

    @material_property
    def density(self):
        """
        Compute the density of the composite based on the molar volumes and masses
        Aliased with self.rho
        """
        densities = np.array([phase.density for phase in self.phases])
        volumes = np.array(
            [phase.molar_volume * molar_fraction for (phase, molar_fraction) in zip(self.phases, self.molar_fractions)])
        return self.averaging_scheme.average_density(volumes, densities)

    @material_property
    def molar_entropy(self):
        """
        Returns enthalpy of the mineral [J]
        Aliased with self.S
        """
        S = sum(phase.molar_entropy * molar_fraction for (
                phase, molar_fraction) in zip(self.phases, self.molar_fractions))
        return S

    @material_property
    def molar_enthalpy(self):
        """
        Returns enthalpy of the mineral [J]
        Aliased with self.H
        """
        H = sum(phase.molar_enthalpy * molar_fraction for (
                phase, molar_fraction) in zip(self.phases, self.molar_fractions))
        return H

    @material_property
    def isothermal_bulk_modulus(self):
        """
        Returns isothermal bulk modulus of the composite [Pa]
        Aliased with self.K_T
        """
        V_frac = np.array([phase.molar_volume * molar_fraction for (
                           phase, molar_fraction) in zip(self.phases, self.molar_fractions)])
        K_ph = np.array(
            [phase.isothermal_bulk_modulus for phase in self.phases])
        G_ph = np.array([phase.shear_modulus for phase in self.phases])

        return self.averaging_scheme.average_bulk_moduli(V_frac, K_ph, G_ph)

    @material_property
    def adiabatic_bulk_modulus(self):
        """
        Returns adiabatic bulk modulus of the mineral [Pa]
        Aliased with self.K_S
        """
        V_frac = np.array([phase.molar_volume * molar_fraction for (
                           phase, molar_fraction) in zip(self.phases, self.molar_fractions)])
        K_ph = np.array(
            [phase.adiabatic_bulk_modulus for phase in self.phases])
        G_ph = np.array([phase.shear_modulus for phase in self.phases])

        return self.averaging_scheme.average_bulk_moduli(V_frac, K_ph, G_ph)

    @material_property
    def isothermal_compressibility(self):
        """
        Returns isothermal compressibility of the composite (or inverse isothermal bulk modulus) [1/Pa]
        Aliased with self.beta_T
        """
        return 1. / self.isothermal_bulk_modulus

    @material_property
    def adiabatic_compressibility(self):
        """
        Returns isothermal compressibility of the composite (or inverse isothermal bulk modulus) [1/Pa]
        Aliased with self.beta_S
        """
        return 1. / self.adiabatic_bulk_modulus

    @material_property
    def shear_modulus(self):
        """
        Returns shear modulus of the mineral [Pa]
        Aliased with self.G
        """
        V_frac = np.array([phase.molar_volume * molar_fraction for (
                           phase, molar_fraction) in zip(self.phases, self.molar_fractions)])
        K_ph = np.array(
            [phase.adiabatic_bulk_modulus for phase in self.phases])
        G_ph = np.array([phase.shear_modulus for phase in self.phases])

        return self.averaging_scheme.average_shear_moduli(V_frac, K_ph, G_ph)

    @material_property
    def p_wave_velocity(self):
        """
        Returns P wave speed of the composite [m/s]
        Aliased with self.v_p
        """
        return np.sqrt((self.adiabatic_bulk_modulus + 4. / 3. *
                        self.shear_modulus) / self.density)

    @material_property
    def bulk_sound_velocity(self):
        """
        Returns bulk sound speed of the composite [m/s]
        Aliased with self.v_phi
        """
        return np.sqrt(self.adiabatic_bulk_modulus / self.density)

    @material_property
    def shear_wave_velocity(self):
        """
        Returns shear wave speed of the composite [m/s]
        Aliased with self.v_s
        """
        return np.sqrt(self.shear_modulus / self.density)

    @material_property
    def grueneisen_parameter(self):
        """
        Returns grueneisen parameter of the composite [unitless]
        Aliased with self.gr
        """
        return self.thermal_expansivity * self.isothermal_bulk_modulus * self.molar_volume / self.molar_heat_capacity_v

    @material_property
    def thermal_expansivity(self):
        """
        Returns thermal expansion coefficient of the composite [1/K]
        Aliased with self.alpha
        """
        volumes = np.array(
            [phase.molar_volume * molar_fraction for (phase, molar_fraction) in zip(self.phases, self.molar_fractions)])
        alphas = np.array([phase.thermal_expansivity for phase in self.phases])
        return self.averaging_scheme.average_thermal_expansivity(volumes, alphas)

    @material_property
    def molar_heat_capacity_v(self):
        """
        Returns molar_heat capacity at constant volume of the composite [J/K/mol]
        Aliased with self.C_v
        """
        c_v = np.array([phase.molar_heat_capacity_v for phase in self.phases])
        return self.averaging_scheme.average_heat_capacity_v(self.molar_fractions, c_v)

    @material_property
    def molar_heat_capacity_p(self):
        """
        Returns molar heat capacity at constant pressure of the composite [J/K/mol]
        Aliased with self.C_p
        """
        c_p = np.array([phase.molar_heat_capacity_p for phase in self.phases])
        return self.averaging_scheme.average_heat_capacity_p(self.molar_fractions, c_p)

    @material_property
    def endmember_partial_gibbs(self):
        """
        Returns the partial Gibbs energies for all
        the endmember minerals in the Composite
        """
        partial_gibbs = np.empty(self.n_endmembers)
        j = 0
        for i, n_endmembers in enumerate(self.endmembers_per_phase):
            if n_endmembers == 1:
                partial_gibbs[j] = self.phases[i].gibbs
            else:
                partial_gibbs[j:j+n_endmembers] = self.phases[i].partial_gibbs
            j += n_endmembers
        return partial_gibbs

    @material_property
    def reaction_affinities(self):
        """
        Returns the affinities corresponding to each reaction in reaction_basis
        """
        return self.reaction_basis.dot(self.endmember_partial_gibbs)

    @material_property
    def equilibrated(self):
        """
        Returns True if the reaction affinities are all zero
        within a given tolerance given by self.equilibrium_tolerance.
        """
        return np.all(np.abs(self.reaction_affinities)
                      < self.equilibrium_tolerance)

    def _mass_to_molar_fractions(self, phases, mass_fractions):
        """
        Converts a set of mass fractions for phases into a set of molar fractions.

        Parameters
        ----------
        phases : list of :class:`burnman.Material`
        The list of phases for which fractions should be converted.

        mass_fractions : numpy array of floats
        An array of mass fractions of the input phases.

        Returns
        -------
        molar_fractions : numpy array of floats
        An array of molar fractions corresponding to the input molar fractions
        """
        molar_masses = np.array([phase.molar_mass for phase in phases])
        moles = mass_fractions / molar_masses
        return moles / sum(moles)

    @cached_property
    def stoichiometric_matrix(self):
        """
        An sympy Matrix where each element M[i,j] corresponds
        to the number of atoms of element[j] in endmember[i].
        """
        def f(i, j):
            e = self.elements[j]
            if e in self.endmember_formulae[i]:
                return nsimplify(self.endmember_formulae[i][e])
            else:
                return 0
        return Matrix(self.n_endmembers, self.n_elements, f)

    @cached_property
    def stoichiometric_array(self):
        """
        An array where each element arr[i,j] corresponds
        to the number of atoms of element[j] in endmember[i].
        """
        return np.array(self.stoichiometric_matrix).astype(float)

    @cached_property
    def reaction_basis(self):
        """
        An array where each element arr[i,j] corresponds
        to the number of moles of endmember[j] involved in reaction[i].
        """
        reaction_basis = np.array([v[:] for v in
                                   self.stoichiometric_matrix.T.nullspace()],
                                  dtype=float)

        if len(reaction_basis) == 0:
            reaction_basis = np.empty((0, self.n_endmembers))

        return reaction_basis

    @cached_property
    def reaction_basis_as_strings(self):
        """
        Returns a list of string representations of all the reactions in
        reaction_basis.
        """
        reaction_strings = []
        for reaction in self.reaction_basis:
            lhs, rhs = ('', '')
            for i, coefficient in enumerate(reaction):
                if coefficient < -1.e-10:
                    if len(lhs) > 0:
                        lhs += ' + '
                    lhs += f'{-coefficient} {self.endmember_names[i]}'
                if coefficient > 1.e-10:
                    if len(rhs) > 0:
                        rhs += ' + '
                    rhs += f'{coefficient} {self.endmember_names[i]}'
            reaction_strings.append(f'{lhs} = {rhs}')

        return reaction_strings

    @cached_property
    def n_reactions(self):
        """
        The number of reactions in reaction_basis.
        """
        return len(self.reaction_basis[:, 0])

    @cached_property
    def independent_element_indices(self):
        """
        A list of an independent set of element indices. If the amounts of
        these elements are known (element_amounts),
        the amounts of the other elements can be
        inferred by -compositional_null_basis[independent_element_indices].dot(element_amounts)
        """
        return sorted(independent_row_indices(self.stoichiometric_matrix.T))

    @cached_property
    def dependent_element_indices(self):
        """
        The element indices not included in the independent list.
        """
        return [i for i in range(self.n_elements)
                if i not in self.independent_element_indices]

    @cached_property
    def reduced_stoichiometric_array(self):
        """
        The stoichiometric array including only the independent elements
        """
        return self.stoichiometric_array[:, self.independent_element_indices]

    @cached_property
    def compositional_null_basis(self):
        """
        An array N such that N.b = 0 for all bulk compositions that can
        be produced with a linear sum of the endmembers in the composite.
        """
        null_basis = np.array([v[:] for v in
                               self.stoichiometric_matrix.nullspace()],
                              dtype=float)

        M = null_basis[:, self.dependent_element_indices]
        assert (M.shape[0] == M.shape[1]) and (M == np.eye(M.shape[0])).all()

        return null_basis

    @cached_property
    def endmember_formulae(self):
        """
        A list of the formulae in the composite.
        """
        self._set_endmember_properties()
        return self.__dict__['endmember_formulae']

    @cached_property
    def endmember_names(self):
        """
        A list of the endmember names contained in the composite.
        Mineral names are returned as given in Mineral.name.
        Solution endmember names are given in the format
        `Mineral.name in SolidSolution.name`.
        """
        self._set_endmember_properties()
        return self.__dict__['endmember_names']

    @cached_property
    def endmembers_per_phase(self):
        """
        A list of integers corresponding to the number of endmembers
        stored within each phase.
        """
        self._set_endmember_properties()
        return self.__dict__['endmembers_per_phase']

    @cached_property
    def elements(self):
        """
        A list of the elements which could be contained in the composite,
        returned in the IUPAC element order.
        """
        self._set_endmember_properties()
        return self.__dict__['elements']

    @cached_property
    def n_endmembers(self):
        """
        Returns the number of endmembers in the composite.
        """
        return len(self.endmember_names)

    @cached_property
    def n_elements(self):
        """
        Returns the total number of distinct elements
        which might be in the composite.
        """
        return len(self.elements)

    def _set_endmember_properties(self):
        """
        Sets endmember_formulae, endmember_names, endmembers_per_phase and
        elements as properties of the Composite. This helper function
        is used to set all properties at the same time while still allowing
        the properties to be stored and documented as individual
        cached_properties.
        """
        endmember_formulae = []
        endmember_names = []
        endmembers_per_phase = []
        for ph_idx, ph in enumerate(self.phases):

            if isinstance(ph, SolidSolution):
                endmember_formulae.extend(ph.endmember_formulae)
                endmember_names.extend([name+' in '+ph.name
                                        for name in ph.endmember_names])
                endmembers_per_phase.append(ph.n_endmembers)

            elif isinstance(ph, Mineral):
                endmember_formulae.append(ph.formula)
                endmember_names.append(ph.name)
                endmembers_per_phase.append(1)

            else:
                raise Exception('Unsupported Material type, can only read'
                                'burnman.Mineral or burnman.SolidSolution')

        # Populate the stoichiometric matrix
        keys = []
        for f in endmember_formulae:
            keys.extend(f.keys())

        # Save to dict so that we only need to do this once
        self.__dict__['endmember_formulae'] = endmember_formulae
        self.__dict__['endmember_names'] = endmember_names
        self.__dict__['endmembers_per_phase'] = endmembers_per_phase
        self.__dict__['elements'] = sort_element_list_to_IUPAC_order(set(keys))



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Sundman 1991 models for iron
Written in the Holland and Powell, 2011 form

EoS terms for bcc are from HP_2011_ds62 for iron
EoS terms for fcc are from an unpublished calibration
(R. Myhill, 02/2017)
"""
from __future__ import absolute_import

from ..classes.mineral import Mineral
from ..classes.solidsolution import SolidSolution
from ..classes.solutionmodel import *
from ..tools.chemistry import dictionarize_formula, formula_mass

"""
ENDMEMBERS
"""


class bcc_iron (Mineral):

    def __init__(self):
        formula = {'Fe': 1.0}
        self.params = {
            'name': 'BCC iron',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': 9149.0,
            'S_0': 36.868,
            'V_0': 7.09e-06,
            'Cp': [21.09, 0.0101455, -221508., 47.1947],
            'a_0': 3.56e-05,
            'K_0': 1.64e+11,
            'Kprime_0': 5.16,
            'Kdprime_0': -3.1e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}
        self.property_modifiers = [
            ['magnetic_chs', {'structural_parameter': 0.4,
                              'curie_temperature': [1043., 0.],
                              'magnetic_moment': [2.22, 0.]}]]
        Mineral.__init__(self)


class fcc_iron (Mineral):

    def __init__(self):
        formula = {'Fe': 1.0}
        self.params = {
            'name': 'FCC iron',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': 7973.0,
            'S_0': 35.907,
            'V_0': 6.93863394593e-06,
            'Cp': [22.24, 0.0088656, -221517., 47.1998],
            'a_0': 5.13e-05,
            'K_0': 1.539e+11,
            'Kprime_0': 5.2,
            'Kdprime_0': -3.37e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}
        self.property_modifiers = [
            ['magnetic_chs', {'structural_parameter': 0.28,
                              'curie_temperature': [201., 0.],
                              'magnetic_moment': [2.1, 0.]}]]
        Mineral.__init__(self)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import inspect
from . import slb
from . import mie_grueneisen_debye as mgd
from . import murnaghan
from . import birch_murnaghan as bm
from . import birch_murnaghan_4th as bm4
from . import modified_tait as mt
from . import dks_liquid
from . import dks_solid
from . import hp
from . import cork
from . import vinet
from . import morse_potential
from . import reciprocal_kprime
from . import aa
from .equation_of_state import EquationOfState


class CombinedMineralMethod(object):
    """Dummy class because CombinedMineral objects are derived
    from a mechanical SolidSolution.
    SolidSolution needs a method to call Mineral.set_state(),
    but a CombinedMineral should never have a method that
    is used for solid solutions."""

    def validate_parameters(self, params):
        pass
    pass


def create(method):
    """
    Creates an instance of an EquationOfState from a string,
    a class EquationOfState, or an instance of EquationOfState.
    """
    if isinstance(method, str):
        if method == "slb2":
            return slb.SLB2()
        elif method == "vinet":
            return vinet.Vinet()
        elif method == "morse":
            return morse_potential.Morse()
        elif method == "rkprime":
            return reciprocal_kprime.RKprime()
        elif method == "aa":
            return aa.AA()
        elif method == "mgd2":
            return mgd.MGD2()
        elif method == "mgd3":
            return mgd.MGD3()
        elif method == "slb3":
            return slb.SLB3()
        elif method == "murnaghan":
            return murnaghan.Murnaghan()
        elif method == "bm2":
            return bm.BM2()
        elif method == "bm3":
            return bm.BM3()
        elif method == "bm4":
            return bm4.BM4()
        elif method == "mt":
            return mt.MT()
        elif method == "hp98":
            return hp.HP98()
        elif method == "hp_tmt":
            return hp.HP_TMT()
        elif method == "hp_tmtL":
            return hp.HP_TMTL()
        elif method == "cork":
            return cork.CORK()
        elif method == "dks_l":
            return dks_liquid.DKS_L()
        elif method == "dks_s":
            return dks_solid.DKS_S()
        elif method == "combined":
            return CombinedMineralMethod()
        else:
            raise Exception("unsupported material method " + method)
    elif isinstance(method, EquationOfState):
        return method
    elif inspect.isclass(method) and issubclass(method, EquationOfState):
        return method()
    else:
        raise Exception(
            "unsupported material method " + method.__class__.__name__)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Murakami_etal_2012
^^^^^^^^^^^^^^^^^^

Minerals from Murakami et al. (2012) supplementary table 5 and references therein, V_0 from
Stixrude & Lithgow-Bertolloni 2005. Some information from personal communication with Murakami.


"""
from __future__ import absolute_import
from ..classes import mineral_helpers as helpers
from ..classes.mineral import Mineral


class mg_perovskite(Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'slb2',
            'V_0': 24.45e-6,  # S & L-B 2005
            'K_0': 281e9,
            'Kprime_0': 4.1,
            'G_0': 173e9,
            'Gprime_0': 1.56,
            'molar_mass': .100,
            'n': 5,
            'Debye_0': 1070.,
            'grueneisen_0': 1.48,
            'q_0': 1.4,
            'eta_s_0': 2.4}

        Mineral.__init__(self)


class mg_perovskite_3rdorder(Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'slb3',
            'V_0': 24.45e-6,  # S & L-B 2005
            'K_0': 281e9,
            'Kprime_0': 4.1,
            'G_0': 171.42e9,
            'Gprime_0': 1.83,
            'molar_mass': .100,
            'n': 5,
            'Debye_0': 1070.,
            'grueneisen_0': 1.48,
            'q_0': 1.4,
            'eta_s_0': 2.4}

        Mineral.__init__(self)


class fe_perovskite(Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'slb2',
            'V_0': 24.607e-6,
            'K_0': 251.9e9,
            'Kprime_0': 4.01,
            'G_0': 164.7e9,
            'Gprime_0': 1.58,
            'molar_mass': .102,
            'n': 5,
            'Debye_0': 1054.,
            'grueneisen_0': 1.48,
            'q_0': 1.4,
            'eta_s_0': 2.4}

        Mineral.__init__(self)


class mg_periclase(Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'slb2',
            'V_0': 11.24e-6,  # S & L-B 2005
            'K_0': 161e9,
            'Kprime_0': 3.9,
            'G_0': 131e9,
            'Gprime_0': 1.92,
            'molar_mass': .0403,
            'n': 2,
            'Debye_0': 773.,  # S& L-B 2005
            'grueneisen_0': 1.5,
            'q_0': 1.5,  # S&L-B 2005
            'eta_s_0': 3.0}

        Mineral.__init__(self)


class fe_periclase(helpers.HelperSpinTransition):

    def __init__(self):
        helpers.HelperSpinTransition.__init__(
            self, 63.0e9, fe_periclase_LS(), fe_periclase_HS())


class fe_periclase_3rd(helpers.HelperSpinTransition):

    def __init__(self):
        helpers.HelperSpinTransition.__init__(
            self, 63.0e9, fe_periclase_LS(), fe_periclase_HS())


class fe_periclase_HS(Mineral):  # From Murakami's emails, see Cayman for details, represents Mg# = .79

    def __init__(self):
        self.params = {
            'equation_of_state': 'slb2',
            'V_0': 11.412e-6,
            'K_0': 159.1e9,
            'Kprime_0': 4.11,
            'G_0': 105.43e9,
            'Gprime_0': 1.773,
            'molar_mass': .047,
            'n': 2,
            'Debye_0': 706.,
            'grueneisen_0': 1.45,
            'q_0': 1.5,
            'eta_s_0': 2.54}

        Mineral.__init__(self)


class fe_periclase_LS(Mineral):  # From Murakami's emails, see Cayman for details, represents Mg# = .79

    def __init__(self):
        self.params = {
            'equation_of_state': 'slb2',
            'V_0': 11.171e-6,
            'K_0': 170.0e9,
            'Kprime_0': 4.00,
            'G_0': 116.34e9,
            'Gprime_0': 1.668,
            'molar_mass': .047,
            'n': 2,
            'Debye_0': 706.,
            'grueneisen_0': 1.45,
            'q_0': 1.5,
            'eta_s_0': 2.54}

        Mineral.__init__(self)


class fe_periclase_HS_3rd(Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'slb3',
            'V_0': 11.412e-6,
            'K_0': 159.1e9,
            'Kprime_0': 4.11,
            'G_0': 129.35e9,
            'Gprime_0': 1.993,
            'molar_mass': .0469,
            'n': 2,
            'Debye_0': 706.,
            'grueneisen_0': 1.45,
            'q_0': 1.5,
            'eta_s_0': 2.54}

        Mineral.__init__(self)


class fe_periclase_LS_3rd(Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'slb3',
            'V_0': 11.171e-6,
            'K_0': 170.0e9,
            'Kprime_0': 4.0,
            'G_0': 151.67e9,
            'Gprime_0': 1.754,
            'molar_mass': .0469,
            'n': 2,
            'Debye_0': 706.,
            'grueneisen_0': 1.45,
            'q_0': 1.5,
            'eta_s_0': 2.54}

        Mineral.__init__(self)


mg_bridgmanite = mg_perovskite
fe_bridgmanite = fe_perovskite
mg_bridgmanite_3rdorder = mg_perovskite_3rdorder
